/*
  Copyright (C) 2012  Torsten Brischalle

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/
*/

#include "lua_func_math_ex.h"

#include "lauxlib.h"
#include "lua_helper.h"
#include "fft.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

//-----------------------------------------------------------------------------

int lua_func_lin_reg(lua_State *lua)
{
  luaL_checktype(lua, 1, LUA_TTABLE);
  luaL_checktype(lua, 2, LUA_TTABLE);
  
  int     cnt = lua_objlen(lua, 1);

  if (cnt < 2)
    lua_throw_error_from_c(lua, "linear regression needs at least two points, table has only %d points", cnt);
  
  if (cnt != (int)lua_objlen(lua, 2))
    lua_throw_error_from_c(lua, "the number of points must be equal for linear regression (table 1 length : %d, table 2 length: %d)", cnt, lua_objlen(lua, 2));
  
  lua_Number  sum_x = 0,
              sum_y = 0,
              sum_xx = 0,
              sum_yy = 0,
              sum_xy = 0;
          
  int     i;
  
  for (i = 1; i <= cnt; i++)
  {
    lua_rawgeti(lua, 1, i);
    lua_rawgeti(lua, 2, i);
    
    lua_Number x = luaL_checknumber(lua, -2);
    lua_Number y = luaL_checknumber(lua, -1);
    
    lua_pop(lua, 2);
    
    sum_x  += x;
    sum_y  += y;
    sum_xx += x * x;
    sum_yy += y * y;
    sum_xy += x * y;
  }
  
  lua_Number v1, v2, v3;
  
  v1 = sum_xx * cnt - sum_x * sum_x;
  if (v1 == 0)
    lua_throw_error_from_c(lua, "invalid values, can not calculate linear regression");
  
  lua_Number b = (sum_xy  * cnt - sum_x * sum_y) / v1;
  lua_Number a = (sum_y - b * sum_x) / cnt;
  
  v1 = sum_xx - sum_x * sum_x / cnt;
  v2 = sum_yy - sum_y * sum_y / cnt;
  
  if ((v1 < 0) || (v2 < 0))
    lua_throw_error_from_c(lua, "invalid values, can not calculate linear regression");
  
  lua_Number corr;
  
  v3 = sqrt(v1) * sqrt(v2);
  if (v3 == 0)
    corr = 1;
  else
    corr = (sum_xy - sum_x * sum_y / cnt) / v3;
  
  lua_pushnumber(lua, a);
  lua_pushnumber(lua, b);
  lua_pushnumber(lua, corr);
  
  return 3;
}

//-----------------------------------------------------------------------------

int lua_func_calc_fft(lua_State *lua)
{
  static float  *float_buf = 0;
  static int    num_points = -1;
  
  int         num = luaL_checkint(lua, 1);
  lua_Number  freq = luaL_checknumber(lua, 2);
  int         pow2;
  int         valid = 0;
  
  luaL_checktype(lua, 3, LUA_TTABLE);
  
  for (pow2 = 2; pow2 <= 17; pow2++)
  {
    if (num == (1 << pow2))
    {
      valid = 1;
      break;
    }
  }
  
  if (!valid)
    lua_throw_error_from_c(lua, "invalid number of FFT points (must be a power of 2 from 2^2=4 to 2^17=131072");
  
  if (num != num_points)
  {
    if (float_buf)
      free(float_buf);
    
    float_buf = (float *)malloc(num * 2 * sizeof(float));
    if (!float_buf)
    {
      num_points = -1;
      lua_throw_error_from_c(lua, "can not allocate buffer for FFT data");
    }
    
    num_points = num;
  }
  
  int i = 0;
  
  for (lua_pushnil(lua); (i < num_points) && lua_next(lua, 3); lua_pop(lua, 1))
  {
    float_buf[(i << 1)] = (float)luaL_checknumber(lua, -1);
    float_buf[(i << 1) + 1] = 0;
    i++;
  }
  
  for (; i < num_points; i++)
  {
    float_buf[(i << 1)] = 0;
    float_buf[(i << 1) + 1] = 0;
  }
  
  fft(float_buf, num_points, 1);
  
  lua_newtable(lua);
  for (i = 0; i < (num_points >> 1); i++)
  {
    int   m = i ? 2 : 1;
    float cr = float_buf[(i << 1)];
    float ci = float_buf[(i << 1) + 1];
    float a = (float)(sqrt(cr * cr + ci * ci) * m / num_points);
    lua_pushnumber(lua, a);
    lua_rawseti(lua, -2, i + 1);
  }
  
  lua_newtable(lua);
  for (i = 0; i < (num_points >> 1); i++)
  {
    float f = (float)i / (float)num_points * (float)freq;
    lua_pushnumber(lua, f);
    lua_rawseti(lua, -2, i + 1);
  }
  
  return 2;
}

//-----------------------------------------------------------------------------

static int lua_func_median_filter_cmp_double(const void *a, const void *b);
static int lua_func_median_filter_cmp_float(const void *a, const void *b);

int lua_func_median_filter_cmp_double(const void *a, const void *b)
{
  if (*(double *)a < *(double *)b)
    return -1;
  else if (*(double *)a > *(double *)b)
    return 1;
  else
    return 0;
}

int lua_func_median_filter_cmp_float(const void *a, const void *b)
{
  if (*(float *)a < *(float *)b)
    return -1;
  else if (*(float *)a > *(float *)b)
    return 1;
  else
    return 0;
}

//-----------------------------------------------------------------------------

int lua_func_median_filter(lua_State *lua)
{
  luaL_checktype(lua, 1, LUA_TTABLE);
  int depth = luaL_checkinteger(lua, 2);
  
  lua_newtable(lua);
  
  int cnt = lua_objlen(lua, 1);

  if (depth < 1)
    lua_throw_error_from_c(lua, "invalid filter depth (must be greater than 0)");
  
  if (depth > cnt)
    lua_throw_error_from_c(lua, "invalid filter depth (must be less than number of points)");
  
  if (depth % 2 == 0)
    lua_throw_error_from_c(lua, "invalid filter depth (must be odd)");
  
  if (cnt < 1)
    return 1;
  
  int         i;
  int         d2 = depth / 2;
  lua_Number  *buf = (lua_Number *)malloc(depth * sizeof(lua_Number));
  
  if (!buf)
    lua_throw_error_from_c(lua, "can not allocate memory (%d bytes)", depth * sizeof(lua_Number));
  
  for (i = 1; i <= cnt; i++)
  {
    if ((i - d2 < 1) || (i + d2 > cnt))
    {
      lua_rawgeti(lua, 1, i);
      lua_rawseti(lua, -2, i);
    }
    else
    {
      int j;
      
      for (j = 0; j < depth; j++)
      {
        lua_rawgeti(lua, 1, i - d2 + j);
        buf[j] = luaL_checknumber(lua, -1);
        lua_pop(lua, 1);
      }
      
      if (sizeof(lua_Number) == sizeof(double))
        qsort(buf, depth, sizeof(double), lua_func_median_filter_cmp_double);
      else
        qsort(buf, depth, sizeof(float), lua_func_median_filter_cmp_float);
        
      lua_pushnumber(lua, buf[d2]);
      lua_rawseti(lua, -2, i);
    }    
  }
  
  free(buf);
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_mean_filter(lua_State *lua)
{
  luaL_checktype(lua, 1, LUA_TTABLE);
  int depth = luaL_checkinteger(lua, 2);
  
  lua_newtable(lua);
  
  int cnt = lua_objlen(lua, 1);

  if (depth < 1)
    lua_throw_error_from_c(lua, "invalid filter depth (must be greater than 0)");
  
  if (depth > cnt)
    lua_throw_error_from_c(lua, "invalid filter depth (must be less than number of points)");
  
  if (depth % 2 == 0)
    lua_throw_error_from_c(lua, "invalid filter depth (must be odd)");
  
  if (cnt < 1)
    return 1;
  
  int     i;
  int     d2 = depth / 2;
  
  for (i = 1; i <= cnt; i++)
  {
    lua_Number  sum = 0;
    int         c = 0;
    int         j;
    
    for (j = 0; j < depth; j++)
    {
      int p = i - d2 + j;
      
      if ((p >= 1) && (p <= cnt))
      {
        lua_rawgeti(lua, 1, p);
        sum += luaL_checknumber(lua, -1);
        lua_pop(lua, 1);
        c++;
      }
    }
    
    lua_pushnumber(lua, sum / c);
    lua_rawseti(lua, -2, i);
  }
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_mean(lua_State *lua)
{
  luaL_checktype(lua, 1, LUA_TTABLE);
   
  int         cnt = 0;
  lua_Number  mean = 0;
  
  for (lua_pushnil(lua); lua_next(lua, 1); lua_pop(lua, 1))
  {
    mean += luaL_checknumber(lua, -1);
    cnt++;
  }
  
  lua_pushnumber(lua, mean / cnt);
  
  return 1;
}

//-----------------------------------------------------------------------------
