/*
  Copyright (C) 2012  Torsten Brischalle

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/
*/

#include "luathread.h"

#include <QMetaType>
#include <iostream>

#include "lualib.h"
#include "lauxlib.h"
#include "lstate.h"
#include "lua_helper.h"
#include "lua_func_sys.h"
#include "globals.h"
#include "register_lua_functions.h"
#include "autolock.h"

using namespace std;

//-----------------------------------------------------------------------------

LuaThread::LuaThread(QObject *tab, QObject *gfx, QObject *dia, QObject *vecgfx)
: QThread(0)
, m_currentLine(-1)
, m_debuggerCmdSemaphore(1)
, m_stdInputMutex(QMutex::Recursive)
{
  m_lua = lua_init();
  m_lua->iGui = this;

  register_lua_functions(m_lua);

  qRegisterMetaType<QVector<double> >("QVector<double>");
  qRegisterMetaType<QList<QString> >("QList<QString>");
  qRegisterMetaType<QPolygon>("QPolygon");
  qRegisterMetaType<QPolygonF>("QPolygonF");

  connect(this, SIGNAL(SignalTableClear()), tab, SLOT(SlotTableClear()));
  connect(this, SIGNAL(SignalTableSetSize(int, int)), tab, SLOT(SlotTableSetSize(int, int)));
  connect(this, SIGNAL(SignalTableSetValue(int, int, const QString &)), tab, SLOT(SlotTableSetValue(int, int, const QString &)));
  connect(this, SIGNAL(SignalTableSetValuesInRow(int, const QList <QString> &)), tab, SLOT(SlotTableSetValuesInRow(int, const QList <QString> &)));
  connect(this, SIGNAL(SignalTableSetValuesInColumn(int, const QList <QString> &)), tab, SLOT(SlotTableSetValuesInColumn(int, const QList <QString> &)));
  connect(this, SIGNAL(SignalTableSaveCsv(const QString &)), tab, SLOT(SlotTableSaveCsv(const QString &)));

  connect(this, SIGNAL(SignalGfxClear()), gfx, SLOT(SlotGfxClear()));
  connect(this, SIGNAL(SignalGfxSetSize(int, int)), gfx, SLOT(SlotGfxSetSize(int, int)));
  connect(this, SIGNAL(SignalGfxDrawPoint(int, int)), gfx, SLOT(SlotGfxDrawPoint(int, int)));
  connect(this, SIGNAL(SignalGfxDrawLine(int, int, int, int)), gfx, SLOT(SlotGfxDrawLine(int, int, int, int)));
  connect(this, SIGNAL(SignalGfxDrawCircle(int, int, int)), gfx, SLOT(SlotGfxDrawCircle(int, int, int)));
  connect(this, SIGNAL(SignalGfxDrawEllipse(int, int, int, int)), gfx, SLOT(SlotGfxDrawEllipse(int, int, int, int)));
  connect(this, SIGNAL(SignalGfxDrawRect(int, int, int, int)), gfx, SLOT(SlotGfxDrawRect(int, int, int, int)));
  connect(this, SIGNAL(SignalGfxDrawPolygon(const QPolygon &)), gfx, SLOT(SlotGfxDrawPolygon(const QPolygon &)));
  connect(this, SIGNAL(SignalGfxFillCircle(int, int, int)), gfx, SLOT(SlotGfxFillCircle(int, int, int)));
  connect(this, SIGNAL(SignalGfxFillEllipse(int, int, int, int)), gfx, SLOT(SlotGfxFillEllipse(int, int, int, int)));
  connect(this, SIGNAL(SignalGfxFillRect(int, int, int, int)), gfx, SLOT(SlotGfxFillRect(int, int, int, int)));
  connect(this, SIGNAL(SignalGfxFillPolygon(const QPolygon &)), gfx, SLOT(SlotGfxFillPolygon(const QPolygon &)));
  connect(this, SIGNAL(SignalGfxDrawText(int, int, const QString &)), gfx, SLOT(SlotGfxDrawText(int, int, const QString &)));
  connect(this, SIGNAL(SignalGfxSetColor(int, int, int)), gfx, SLOT(SlotGfxSetColor(int, int, int)));
  connect(this, SIGNAL(SignalGfxSetColor(const QString &)), gfx, SLOT(SlotGfxSetColor(const QString &)));
  connect(this, SIGNAL(SignalGfxSetFillColor(int, int, int)), gfx, SLOT(SlotGfxSetFillColor(int, int, int)));
  connect(this, SIGNAL(SignalGfxSetFillColor(const QString &)), gfx, SLOT(SlotGfxSetFillColor(const QString &)));
  connect(this, SIGNAL(SignalGfxBeginUpdate()), gfx, SLOT(SlotGfxBeginUpdate()));
  connect(this, SIGNAL(SignalGfxEndUpdate()), gfx, SLOT(SlotGfxEndUpdate()));
  connect(this, SIGNAL(SignalGfxSaveImg(const QString &)), gfx, SLOT(SlotGfxSaveImg(const QString &)));

  connect(this, SIGNAL(SignalDiaDiagramNew(const QString &)), dia, SLOT(SlotDiaDiagramNew(const QString &)));
  connect(this, SIGNAL(SignalDiaDiagramDelete(const QString &)), dia, SLOT(SlotDiaDiagramDelete(const QString &)));
  connect(this, SIGNAL(SignalDiaDiagramSaveImg(const QString &, const QString &)), dia, SLOT(SlotDiaDiagramSaveImg(const QString &, const QString &)));
  connect(this, SIGNAL(SignalDiaDiagramSaveCsv(const QString &, const QString &)), dia, SLOT(SlotDiaDiagramSaveCsv(const QString &, const QString &)));
  connect(this, SIGNAL(SignalDiaDiagramDeleteAll()), dia, SLOT(SlotDiaDiagramDeleteAll()));
  connect(this, SIGNAL(SignalDiaCurveNew(const QString &, const QString &, int)), dia, SLOT(SlotDiaCurveNew(const QString &, const QString &, int)));
  connect(this, SIGNAL(SignalDiaCurveDelete(const QString &, const QString &)), dia, SLOT(SlotDiaCurveDelete(const QString &, const QString &)));
  connect(this, SIGNAL(SignalDiaCurveDeleteAll(const QString &)), dia, SLOT(SlotDiaCurveDeleteAll(const QString &)));
  connect(this, SIGNAL(SignalDiaCurveClearData(const QString &, const QString &)), dia, SLOT(SlotDiaCurveClearData(const QString &, const QString &)));
  connect(this, SIGNAL(SignalDiaCurveSetData(const QString &, const QString &, const QVector <double> &, const QVector <double> &)), dia, SLOT(SlotDiaCurveSetData(const QString &, const QString &, const QVector <double> &, const QVector <double> &)));
  connect(this, SIGNAL(SignalDiaCurveAppendData(const QString &, const QString &, double, double)), dia, SLOT(SlotDiaCurveAppendData(const QString &, const QString &, double, double)));

  connect(this, SIGNAL(SignalVecGfxClear()), vecgfx, SLOT(SlotVecGfxClear()));
  connect(this, SIGNAL(SignalVecGfxDrawLine(qreal, qreal, qreal, qreal)), vecgfx, SLOT(SlotVecGfxDrawLine(qreal, qreal, qreal, qreal)));
  connect(this, SIGNAL(SignalVecGfxDrawCircle(qreal, qreal, qreal)), vecgfx, SLOT(SlotVecGfxDrawCircle(qreal, qreal, qreal)));
  connect(this, SIGNAL(SignalVecGfxDrawEllipse(qreal, qreal, qreal, qreal)), vecgfx, SLOT(SlotVecGfxDrawEllipse(qreal, qreal, qreal, qreal)));
  connect(this, SIGNAL(SignalVecGfxDrawRect(qreal, qreal, qreal, qreal)), vecgfx, SLOT(SlotVecGfxDrawRect(qreal, qreal, qreal, qreal)));
  connect(this, SIGNAL(SignalVecGfxDrawPolygon(const QPolygonF &)), vecgfx, SLOT(SlotVecGfxDrawPolygon(const QPolygonF &)));
  connect(this, SIGNAL(SignalVecGfxFillCircle(qreal, qreal, qreal)), vecgfx, SLOT(SlotVecGfxFillCircle(qreal, qreal, qreal)));
  connect(this, SIGNAL(SignalVecGfxFillEllipse(qreal, qreal, qreal, qreal)), vecgfx, SLOT(SlotVecGfxFillEllipse(qreal, qreal, qreal, qreal)));
  connect(this, SIGNAL(SignalVecGfxFillRect(qreal, qreal, qreal, qreal)), vecgfx, SLOT(SlotVecGfxFillRect(qreal, qreal, qreal, qreal)));
  connect(this, SIGNAL(SignalVecGfxFillPolygon(const QPolygonF &)), vecgfx, SLOT(SlotVecGfxFillPolygon(const QPolygonF &)));
  connect(this, SIGNAL(SignalVecGfxDrawText(qreal, qreal, const QString &)), vecgfx, SLOT(SlotVecGfxDrawText(qreal, qreal, const QString &)));
  connect(this, SIGNAL(SignalVecGfxSetColor(int, int, int)), vecgfx, SLOT(SlotVecGfxSetColor(int, int, int)));
  connect(this, SIGNAL(SignalVecGfxSetColor(const QString &)), vecgfx, SLOT(SlotVecGfxSetColor(const QString &)));
  connect(this, SIGNAL(SignalVecGfxSetFillColor(int, int, int)), vecgfx, SLOT(SlotVecGfxSetFillColor(int, int, int)));
  connect(this, SIGNAL(SignalVecGfxSetFillColor(const QString &)), vecgfx, SLOT(SlotVecGfxSetFillColor(const QString &)));
  connect(this, SIGNAL(SignalVecGfxSaveImg(const QString &, int, int)), vecgfx, SLOT(SlotVecGfxSaveImg(const QString &, int, int)));
  connect(this, SIGNAL(SignalVecGfxSaveSvg(const QString &)), vecgfx, SLOT(SlotVecGfxSaveCsv(const QString &)));

  m_debuggerHalted = false;
  m_debuggerCmd = DEBUGGER_CMD_NONE;
  m_luaRunning = false;
  m_stackDepth = -1;

  ClearBreakpoints();
}

//-----------------------------------------------------------------------------

LuaThread::~LuaThread()
{
  if (m_lua)
    lua_close(m_lua);
}

//-----------------------------------------------------------------------------

void LuaThread::Execute(const QString & luaCode)
{
  if (isRunning())
    throw runtime_error("LuaThread: can not execute code because lua is running");

  m_luaCode = luaCode;
  ClearStdInput();
  start();
}

//-----------------------------------------------------------------------------

void LuaThread::run()
{
  emit LuaStarted();

  try
  {
    g_lua_stop = 0;
    m_currentLine = -1;
    m_luaRunning = true;
    m_stackDepth = -1;
    m_debuggerCmd = DEBUGGER_CMD_NONE;

    lua_do_string(m_lua, m_luaCode.toAscii());

    m_luaRunning = false;
    m_currentLine = -1;
    m_stackDepth = -1;
    m_debuggerCmd = DEBUGGER_CMD_NONE;
  }
  catch (...)
  {
    // nothing to do on error
  }

  emit LuaStopped();
}

//-----------------------------------------------------------------------------

void LuaThread::TableClear()
{
  emit SignalTableClear();
}

//-----------------------------------------------------------------------------

void LuaThread::TableSetSize(int rows, int columns)
{
  emit SignalTableSetSize(rows, columns);
}

//-----------------------------------------------------------------------------

void LuaThread::TableSetValue(int row, int column, const QString & value)
{
  emit SignalTableSetValue(row, column, value);
}

//-----------------------------------------------------------------------------

void LuaThread::TableSetValuesInRow(int row, const QList<QString> &valueList)
{
  emit SignalTableSetValuesInRow(row, valueList);
}

//-----------------------------------------------------------------------------

void LuaThread::TableSetValuesInColumn(int column, const QList<QString> &valueList)
{
  emit SignalTableSetValuesInColumn(column, valueList);
}

//-----------------------------------------------------------------------------

void LuaThread::TableSaveCsv(const QString &fileName)
{
  emit SignalTableSaveCsv(fileName);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxClear()
{
  emit SignalGfxClear();
}

//-----------------------------------------------------------------------------

void LuaThread::GfxSetSize(int x, int y)
{
  emit SignalGfxSetSize(x, y);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxDrawPoint(int x, int y)
{
  emit SignalGfxDrawPoint(x, y);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxDrawLine(int x1, int y1, int x2, int y2)
{
  emit SignalGfxDrawLine(x1, y1, x2, y2);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxDrawCircle(int x, int y, int r)
{
  emit SignalGfxDrawCircle(x, y, r);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxDrawEllipse(int x, int y, int rx, int ry)
{
  emit SignalGfxDrawEllipse(x, y, rx, ry);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxDrawRect(int x, int y, int w, int h)
{
  emit SignalGfxDrawRect(x, y, w, h);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxDrawPolygon(const QPolygon &polygon)
{
  emit SignalGfxDrawPolygon(polygon);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxFillCircle(int x, int y, int r)
{
  emit SignalGfxFillCircle(x, y, r);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxFillEllipse(int x, int y, int rx, int ry)
{
  emit SignalGfxFillEllipse(x, y, rx, ry);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxFillRect(int x, int y, int w, int h)
{
  emit SignalGfxFillRect(x, y, w, h);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxFillPolygon(const QPolygon &polygon)
{
  emit SignalGfxFillPolygon(polygon);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxDrawText(int x, int y, const QString &text)
{
  emit SignalGfxDrawText(x, y, text);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxSetColor(int red, int green, int blue)
{
  emit SignalGfxSetColor(red, green, blue);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxSetColor(const QString & name)
{
  emit SignalGfxSetColor(name);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxSetFillColor(int red, int green, int blue)
{
  emit SignalGfxSetFillColor(red, green, blue);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxSetFillColor(const QString & name)
{
  emit SignalGfxSetFillColor(name);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxBeginUpdate()
{
  emit SignalGfxBeginUpdate();
}

//-----------------------------------------------------------------------------

void LuaThread::GfxEndUpdate()
{
  emit SignalGfxEndUpdate();
}

//-----------------------------------------------------------------------------

void LuaThread::DiaDiagramNew(const QString & diagramName)
{
  emit SignalDiaDiagramNew(diagramName);
}

//-----------------------------------------------------------------------------

void LuaThread::DiaDiagramDelete(const QString & diagramName)
{
  emit SignalDiaDiagramDelete(diagramName);
}

//-----------------------------------------------------------------------------

void LuaThread::DiaDiagramDeleteAll()
{
  emit SignalDiaDiagramDeleteAll();
}

//-----------------------------------------------------------------------------

void LuaThread::DiaCurveNew(const QString & diagramName, const QString & curveName, int yAxis)
{
  emit SignalDiaCurveNew(diagramName, curveName, yAxis);
}

//-----------------------------------------------------------------------------

void LuaThread::DiaCurveDelete(const QString & diagramName, const QString & curveName)
{
  emit SignalDiaCurveDelete(diagramName, curveName);
}

//-----------------------------------------------------------------------------

void LuaThread::DiaCurveDeleteAll(const QString & diagramName)
{
  emit SignalDiaCurveDeleteAll(diagramName);
}

//-----------------------------------------------------------------------------

void LuaThread::DiaCurveClearData(const QString & diagramName, const QString & curveName)
{
  emit SignalDiaCurveClearData(diagramName, curveName);
}

//-----------------------------------------------------------------------------

void LuaThread::DiaCurveSetData(const QString & diagramName, const QString &curveName, const QVector <double> & x, const QVector <double> & y)
{
  emit SignalDiaCurveSetData(diagramName, curveName, x, y);
}

//-----------------------------------------------------------------------------

void LuaThread::DiaCurveAppendData(const QString & diagramName, const QString &curveName, double x, double y)
{
  emit SignalDiaCurveAppendData(diagramName, curveName, x, y);
}

//-----------------------------------------------------------------------------

void LuaThread::Debug(bool debugOn)
{
  if (!m_debuggerHalted)
  {
    if (debugOn)
    {
      lua_sethook(m_lua, debugger_lua_hook, LUA_MASKLINE, 0);
    }
    else
    {
      lua_sethook(m_lua, 0, 0, 0);
    }
  }
  else
  {
    throw runtime_error("debugger is halted");
  }
}

//-----------------------------------------------------------------------------

void LuaThread::debugger_lua_hook(lua_State *lua, lua_Debug *ar)
{
  LuaThread *luaThread = dynamic_cast <LuaThread *>(lua->iGui);
  if (luaThread)
    luaThread->DebuggerLuaHook(lua, ar);
}

//-----------------------------------------------------------------------------

void LuaThread::DebuggerLuaHook(lua_State *lua, lua_Debug *ar)
{
  (void)lua;

  if (ar->event == LUA_HOOKLINE)
  {
    if (!m_debuggerHalted)
    {
      bool halt = false;

      if (m_debuggerCmd == DEBUGGER_CMD_HALT)
      {
        if ((m_stackDepth == -1) || (m_stackDepth >= GetLuaRuntimeStackDepth()))
          halt = true;
      }

      if (!halt)
      {
        int i;

        for (i = 0; i < DEBUGGER_MAX_BREAKPOINTS; i++)
        {
          if (m_debuggerBreakpoints[i] == -1)
            break;

          if (m_debuggerBreakpoints[i] == ar->currentline)
          {
            halt = true;
            break;
          }
        }
      }

      if (halt)
      {
        m_debuggerCmd = DEBUGGER_CMD_NONE;
        m_debuggerHalted = true;
        m_currentLine = ar->currentline;
        emit LuaHalted();
      }
    }

    while (m_debuggerHalted)
    {
      while (m_debuggerCmdSemaphore.tryAcquire(1, 500) == false)
      {
        if (g_lua_stop)
        {
          m_debuggerCmd = DEBUGGER_CMD_NONE;
          m_debuggerHalted = false;
          break;
        }
      }

      switch (m_debuggerCmd)
      {
      case DEBUGGER_CMD_CONTINUE:
        m_debuggerCmd = DEBUGGER_CMD_NONE;
        m_debuggerHalted = false;
        break;

      case DEBUGGER_CMD_STEP_IN:
        m_stackDepth = -1;
        m_debuggerCmd = DEBUGGER_CMD_HALT;
        m_debuggerHalted = false;
        break;

      case DEBUGGER_CMD_STEP_OUT:
        m_stackDepth = GetLuaRuntimeStackDepth() - 1;
        if (m_stackDepth < 0)
          m_stackDepth = 0;
        m_debuggerCmd = DEBUGGER_CMD_HALT;
        m_debuggerHalted = false;
        break;

      case DEBUGGER_CMD_STEP_OVER:
        m_stackDepth = GetLuaRuntimeStackDepth();
        m_debuggerCmd = DEBUGGER_CMD_HALT;
        m_debuggerHalted = false;
        break;

      default:
        break;
      }
    }

    m_currentLine = -1;
    emit LuaContinued();
  }
}

//-----------------------------------------------------------------------------

void LuaThread::DebuggerHalt()
{
  if (m_luaRunning)
  {
    if (!m_debuggerHalted)
    {
      m_stackDepth = -1;
      m_debuggerCmd = DEBUGGER_CMD_HALT;
    }
  }
  else
    throw runtime_error("not running");
}

//-----------------------------------------------------------------------------

void LuaThread::DebuggerContinue()
{
  if (m_luaRunning)
  {
    if (m_debuggerHalted)
    {
      m_debuggerCmd = DEBUGGER_CMD_CONTINUE;
      m_debuggerCmdSemaphore.release(1);
    }
  }
  else
    throw runtime_error("not running");
}

//-----------------------------------------------------------------------------

void LuaThread::DebuggerStepIn()
{
  if (m_luaRunning)
  {
    if (m_debuggerHalted)
    {
      m_debuggerCmd = DEBUGGER_CMD_STEP_IN;
      m_debuggerCmdSemaphore.release(1);
    }
  }
  else
    throw runtime_error("not running");
}

//-----------------------------------------------------------------------------

void LuaThread::DebuggerStepOut()
{
  if (m_luaRunning)
  {
    if (m_debuggerHalted)
    {
      m_debuggerCmd = DEBUGGER_CMD_STEP_OUT;
      m_debuggerCmdSemaphore.release(1);
    }
  }
  else
    throw runtime_error("not running");
}

//-----------------------------------------------------------------------------

void LuaThread::DebuggerStepOver()
{
  if (m_luaRunning)
  {
    if (m_debuggerHalted)
    {
      m_debuggerCmd = DEBUGGER_CMD_STEP_OVER;
      m_debuggerCmdSemaphore.release(1);
    }
  }
  else
    throw runtime_error("not running");
}

//-----------------------------------------------------------------------------

void LuaThread::ClearBreakpoints()
{
  int bp;

  for (bp = 0; bp < DEBUGGER_MAX_BREAKPOINTS; bp++)
    m_debuggerBreakpoints[bp] = -1;
}

//-----------------------------------------------------------------------------

void LuaThread::SetBreakpoint(int line)
{
  int   i;

  for (i = 0; i < DEBUGGER_MAX_BREAKPOINTS; i++)
  {
    if (m_debuggerBreakpoints[i] == line)
      return;
    else if (m_debuggerBreakpoints[i] == -1)
    {
      m_debuggerBreakpoints[i] = line;
      return;
    }
  }

  throw runtime_error("maximum number of breakpoints reached");
}

//-----------------------------------------------------------------------------

void LuaThread::ClearBreakpoint(int line)
{
  int i;

  for (i = 0; i < DEBUGGER_MAX_BREAKPOINTS; i++)
  {
    if (m_debuggerBreakpoints[i] == line)
    {
      int j;

      for (j = i; j < DEBUGGER_MAX_BREAKPOINTS - 1; j++)
      {
        m_debuggerBreakpoints[j] = m_debuggerBreakpoints[j + 1];
        if (m_debuggerBreakpoints[j] == -1)
          break;
      }

      m_debuggerBreakpoints[DEBUGGER_MAX_BREAKPOINTS - 1] = -1;
      break;
    }
  }
}

//-----------------------------------------------------------------------------

bool LuaThread::IsBreakpoint(int line)
{
  int i;

  for (i = 0; i < DEBUGGER_MAX_BREAKPOINTS; i++)
  {
    if (m_debuggerBreakpoints[i] == line)
      return true;
    else if (m_debuggerBreakpoints[i] == -1)
      return false;
  }

  return false;
}

//-----------------------------------------------------------------------------

bool LuaThread::DebugEnabled()
{
  return m_lua->hook != 0;
}

//-----------------------------------------------------------------------------

int LuaThread::CurrentLine()
{
  return m_currentLine;
}

//-----------------------------------------------------------------------------

int LuaThread::GetLuaRuntimeStackDepth()
{
  lua_Debug ar;
  int       depth = 0;

  while (lua_getstack(m_lua, depth, &ar))
    depth++;

  return depth;
}

//-----------------------------------------------------------------------------

CallStack LuaThread::DebuggerCallStack()
{
  if (m_debuggerHalted)
  {
    CallStack callStack;
    callStack.GetCallStack(m_lua);
    return callStack;
  }
  else
    throw runtime_error("debugger not halted");
}

//-----------------------------------------------------------------------------

VariableList LuaThread::DebuggerGetLocalVariables()
{
  if (m_debuggerHalted)
  {
    VariableList varList;
    varList.GetLocalVariables(m_lua);
    return varList;
  }
  else
    throw runtime_error("debugger not halted");
}

//-----------------------------------------------------------------------------

VariableList LuaThread::DebuggerGetGlobalVariables()
{
  if (m_debuggerHalted)
  {
    VariableList varList;
    varList.GetGlobalVariables(m_lua);
    return varList;
  }
  else
    throw runtime_error("debugger not halted");
}

//-----------------------------------------------------------------------------

VariableList LuaThread::DebuggerGetLocalVariablesFromPath(const QVariantList &path)
{
  if (m_debuggerHalted)
  {
    VariableList varList;
    varList.GetLocalVariablesFromPath(m_lua, path);
    return varList;
  }
  else
    throw runtime_error("debugger not halted");
}

//-----------------------------------------------------------------------------

VariableList LuaThread::DebuggerGetGlobalVariablesFromPath(const QVariantList &path)
{
  if (m_debuggerHalted)
  {
    VariableList varList;
    varList.GetGlobalVariablesFromPath(m_lua, path);
    return varList;
  }
  else
    throw runtime_error("debugger not halted");
}

//-----------------------------------------------------------------------------

void LuaThread::ResetLua()
{
  if (!m_luaRunning)
  {
    if (m_lua)
      lua_close(m_lua);

    m_lua = lua_init();
    m_lua->iGui = this;

    register_lua_functions(m_lua);

    m_debuggerHalted = false;
    m_debuggerCmd = DEBUGGER_CMD_NONE;
    m_luaRunning = false;
    m_stackDepth = -1;

    ClearBreakpoints();
    ClearStdInput();
  }
  else
    throw runtime_error("can not reset Lua because a script is running");
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxClear()
{
  emit SignalVecGfxClear();
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxSetColor(int red, int green, int blue)
{
  emit SignalVecGfxSetColor(red, green, blue);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxSetColor(const QString & name)
{
  emit SignalVecGfxSetColor(name);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxSetFillColor(int red, int green, int blue)
{
  emit SignalVecGfxSetFillColor(red, green, blue);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxSetFillColor(const QString & name)
{
  emit SignalVecGfxSetFillColor(name);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxDrawLine(qreal x1, qreal y1, qreal x2, qreal y2)
{
  emit SignalVecGfxDrawLine(x1, y1, x2, y2);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxDrawCircle(qreal x, qreal y, qreal r)
{
  emit SignalVecGfxDrawCircle(x, y, r);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxDrawEllipse(qreal x, qreal y, qreal rx, qreal ry)
{
  emit SignalVecGfxDrawEllipse(x, y, rx, ry);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxDrawRect(qreal x, qreal y, qreal w, qreal h)
{
  emit SignalVecGfxDrawRect(x, y, w, h);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxDrawPolygon(const QPolygonF &polygon)
{
  emit SignalVecGfxDrawPolygon(polygon);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxFillCircle(qreal x, qreal y, qreal r)
{
  emit SignalVecGfxFillCircle(x, y, r);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxFillEllipse(qreal x, qreal y, qreal rx, qreal ry)
{
  emit SignalVecGfxFillEllipse(x, y, rx, ry);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxFillRect(qreal x, qreal y, qreal w, qreal h)
{
  emit SignalVecGfxFillRect(x, y, w, h);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxFillPolygon(const QPolygonF &polygon)
{
  emit SignalVecGfxFillPolygon(polygon);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxDrawText(qreal x, qreal y, const QString &text)
{
  emit SignalVecGfxDrawText(x, y, text);
}

//-----------------------------------------------------------------------------

void LuaThread::GfxSaveImg(const QString &fileName)
{
  emit SignalGfxSaveImg(fileName);
}

//-----------------------------------------------------------------------------

void LuaThread::DiaDiagramSaveImg(const QString &diagramName, const QString &fileName)
{
  emit SignalDiaDiagramSaveImg(diagramName, fileName);
}

//-----------------------------------------------------------------------------

void LuaThread::DiaDiagramSaveCsv(const QString &diagramName, const QString &fileName)
{
  emit SignalDiaDiagramSaveCsv(diagramName, fileName);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxSaveImg(const QString &fileName, int widht, int height)
{
  emit SignalVecGfxSaveImg(fileName, widht, height);
}

//-----------------------------------------------------------------------------

void LuaThread::VecGfxSaveSvg(const QString &fileName)
{
  emit SignalVecGfxSaveSvg(fileName);
}

//-----------------------------------------------------------------------------

void LuaThread::PutToStdInput(const QString & str)
{
  AutoLock lock(m_stdInputMutex);

  string  s = str.toStdString();
  int     i;
  int     len = (int)s.length();

  for (i = 0; i < len; i++)
    m_stdInput.push_back(s[i]);
}

//-----------------------------------------------------------------------------

void LuaThread::ClearStdInput()
{
  AutoLock lock(m_stdInputMutex);

  m_stdInput.clear();
}

//-----------------------------------------------------------------------------

bool LuaThread::ConsoleTestEof()
{
  AutoLock lock(m_stdInputMutex);

  return (m_stdInput.size() != 0);
}

//-----------------------------------------------------------------------------

int LuaThread::ConsoleReadChars(char *str, int count)
{
  int n = 0;

  while ((n < count) && !g_lua_stop)
  {
    m_stdInputMutex.lock();

    if (m_stdInput.size() != 0)
    {
      str[n] = m_stdInput.front();
      m_stdInput.pop_front();
      n++;
      m_stdInputMutex.unlock();
    }
    else
    {
      m_stdInputMutex.unlock();
      QThreadDelay::Delay(100);
    }
  }

  return n;
}

//-----------------------------------------------------------------------------

bool LuaThread::ConsoleReadLine(char *str, int maxNum)
{
  int n = 0;

  while (!g_lua_stop)
  {
    m_stdInputMutex.lock();

    if (m_stdInput.size() != 0)
    {
      char c = m_stdInput.front();
      m_stdInput.pop_front();
      str[n++] = c;

      m_stdInputMutex.unlock();

      if (c == '\n')
        break;

      if (n == (maxNum - 1))
        break;
    }
    else
    {
      m_stdInputMutex.unlock();
      QThreadDelay::Delay(100);
    }
  }

  str[n] = 0;

  return n > 0;
}

//-----------------------------------------------------------------------------

bool LuaThread::ConsoleReadNumber(double *number)
{
  *number = 0;

  string numberStr;

  while (!g_lua_stop)
  {
    m_stdInputMutex.lock();

    if (m_stdInput.size() != 0)
    {
      numberStr += m_stdInput.front();

      char    *endPtr;
      double  d = strtod(numberStr.c_str(), &endPtr);
      int     cnt = (int)(endPtr - numberStr.c_str());

      if (cnt == (int)numberStr.length())
      {
        m_stdInput.pop_front();
        m_stdInputMutex.unlock();
      }
      else
      {
        bool valid = false;

        if (cnt > 0)
        {
          if (m_stdInput.front() == '\n')
            m_stdInput.pop_front();

          *number = d;
          valid = true;
        }

        m_stdInputMutex.unlock();
        return valid;
      }
    }
    else
    {
      m_stdInputMutex.unlock();
      QThreadDelay::Delay(100);
    }
  }

  return false;
}

//-----------------------------------------------------------------------------
