/*
  modified and extended by Torsten Brischalle

  base upon editlog_stream.h

  Version:	1.0, created 2000-04-02
  Author:	Daniel Lohmann (daniel@uni-koblenz.de)

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/
*/

#ifndef __EDITLOG_STREAM_H__
#define __EDITLOG_STREAM_H__

//-----------------------------------------------------------------------------

#include "qconsolewidget.h"
#include "autolock.h"

//-----------------------------------------------------------------------------

template <typename E, typename T = std::char_traits<E>, int BUF_SIZE = 512 >
class basic_QConsoleStreamBuf : public std::basic_streambuf< E, T >
{
  typedef typename std::basic_streambuf< E, T>::int_type int_type;
  typedef typename std::basic_streambuf< E, T>::char_type char_type;
  typedef typename std::basic_streambuf< E, T>::traits_type traits_type;

public:
  basic_QConsoleStreamBuf(QConsoleWidget *console);
  virtual ~basic_QConsoleStreamBuf();

protected:

  virtual int_type overflow( int_type c = T::eof() );
  virtual int sync();
  virtual std::streamsize xsputn(const char_type* pch, std::streamsize n);

  char_type         *psz;
  QConsoleWidget    *m_console;
  QMutex            m_mutex;
};

//-----------------------------------------------------------------------------

// Ansi version
typedef basic_QConsoleStreamBuf< char >		QConsoleStreamBuf;

// Unicode version
typedef basic_QConsoleStreamBuf< wchar_t>	wQConsoleStreamBuf;

// TCHAR version
#ifdef _UNICODE
#define _tQConsoleStreamBuf wQConsoleStreamBuf
#else
#define _tQConsoleStreamBuf QConsoleStreamBuf
#endif

//-----------------------------------------------------------------------------
// IMPLEMENTATION
//-----------------------------------------------------------------------------

template <typename E, typename T, int BUF_SIZE >
basic_QConsoleStreamBuf<E,T,BUF_SIZE>::basic_QConsoleStreamBuf(QConsoleWidget *console)
: std::basic_streambuf<E,T>()
, m_console(console)
, m_mutex(QMutex::Recursive)
{
  psz = new char_type[ BUF_SIZE ];
  this->setbuf( psz, BUF_SIZE );

  // leave place for single char + 0 terminator
  setp( psz, psz + BUF_SIZE - 2 );
}

//-----------------------------------------------------------------------------

template <typename E, typename T, int BUF_SIZE >
basic_QConsoleStreamBuf<E,T,BUF_SIZE>::~basic_QConsoleStreamBuf()
{
  delete psz;
}

//-----------------------------------------------------------------------------

template <typename E, typename T, int BUF_SIZE >
int basic_QConsoleStreamBuf<E,T,BUF_SIZE>::sync()
{
  overflow();
  return 0;
}

//-----------------------------------------------------------------------------

template <typename E, typename T, int BUF_SIZE >
typename std::streamsize basic_QConsoleStreamBuf<E,T,BUF_SIZE>::xsputn(const char_type* pch, std::streamsize n)
{
  AutoLock lock(m_mutex);

  std::streamsize nMax, nPut;

  for( nPut = 0; 0 < n; )
  {
    if( this->pptr() != 0 && 0 < (nMax = this->epptr() - this->pptr()) )
    {
      if( n < nMax )
        nMax = n;

      traits_type::copy( this->pptr(), pch, nMax );

      // Sync if string contains LF
      bool bSync = traits_type::find( pch, nMax, traits_type::to_char_type( '\n' ) ) != NULL;
      pch += nMax, nPut += nMax, n -= nMax, this->pbump(nMax);
      if( bSync )
        sync();
    }
    else if (traits_type::eq_int_type(traits_type::eof(), overflow(traits_type::to_int_type(*pch))))
      break;
    else
      ++pch, ++nPut, --n;
  }

  return (nPut);
}

//-----------------------------------------------------------------------------

template <typename E, typename T, int BUF_SIZE >
typename basic_QConsoleStreamBuf< E, T, BUF_SIZE >::int_type basic_QConsoleStreamBuf< E,T,BUF_SIZE >::overflow(int_type c)
{
  AutoLock lock(m_mutex);

  char_type* plast = this->pptr();
  if( c != T::eof() )
    *plast++ = c; // add c to buffer

  *plast = char_type();

  // Pass text to the edit control
  m_console->Write( this->pbase());

  setp( this->pbase(), this->epptr() );

  return c != T::eof() ? T::not_eof( c ) : T::eof();
}

//-----------------------------------------------------------------------------

#endif
