/*
  Copyright (C) 2012  Torsten Brischalle

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/
*/

#include "windowmain.h"
#include "ui_windowmain.h"

#include <QMessageBox>
#include <QTextDocumentWriter>
#include <QSettings>
#include <QGraphicsSimpleTextItem>
#include <QtSvg/QSvgGenerator>
#include <QFontDatabase>
#include <QStyleFactory>

#include "globals.h"
#include "qsciscintilla.h"
#include "qscilexerlua.h"

Q_DECLARE_METATYPE(QList<int>)

using namespace std;

//-----------------------------------------------------------------------------

WindowMain::WindowMain(QWidget *parent) :
  QMainWindow(parent),
  ui(new Ui::WindowMain),
  m_actionGroupGuiStyle(this)
{
  ui->setupUi(this);

  m_font = GetFixedFont(11);

  setWindowIcon(QIcon("icons:Lua-Quick-Try-Out.png"));

  m_luaThread = QSharedPointer <LuaThread>(
    new LuaThread(this, ui->gfxView, this, this)
  );

  m_debugInfoWidthDebugOn.push_back(800);
  m_debugInfoWidthDebugOn.push_back(200);
  m_debugInfoWidthDebugOff.push_back(1000);
  m_debugInfoWidthDebugOff.push_back(0);

  connect(m_luaThread.data(), SIGNAL(LuaStarted()), this, SLOT(LuaStarted()));
  connect(m_luaThread.data(), SIGNAL(LuaStopped()), this, SLOT(LuaStopped()));
  connect(m_luaThread.data(), SIGNAL(LuaHalted()), this, SLOT(LuaHalted()));
  connect(m_luaThread.data(), SIGNAL(LuaContinued()), this, SLOT(LuaContinued()));

  ui->console->AttachLuaThread(m_luaThread.data());

  m_consoleStreamBufOut = new QConsoleStreamBuf(ui->console);
  m_consoleStreamBufErr = new QConsoleStreamBuf(ui->console);

  m_oldStreamBufOut = cout.rdbuf(m_consoleStreamBufOut);
  m_oldStreamBufErr = cerr.rdbuf(m_consoleStreamBufErr);

  SetGuiState(GS_STOPPED);

  UpdateWindowTitle();

  ui->tabWidget->setCurrentIndex(0);

  ui->menuBar->insertSeparator(ui->menuHelp->menuAction());

  ui->console->setFont(m_font);
  ui->scintilla->setFont(m_font);

  QsciLexerLua *lexer = new QsciLexerLua();

  lexer->setDefaultFont(m_font);
  lexer->setFont(m_font);

  lexer->setColor(Qt::darkRed, QsciLexerLua::BasicFunctions);
  lexer->setPaper(lexer->defaultPaper(QsciLexerLua::Default));

  lexer->setColor(Qt::darkRed, QsciLexerLua::StringTableMathsFunctions);
  lexer->setPaper(lexer->defaultPaper(QsciLexerLua::Default));

  lexer->setColor(Qt::darkRed, QsciLexerLua::CoroutinesIOSystemFacilities);
  lexer->setPaper(lexer->defaultPaper(QsciLexerLua::Default));

  ui->scintilla->setLexer(lexer);

  ui->scintilla->setFolding(QsciScintilla::BoxedTreeFoldStyle);
  ui->scintilla->setAutoIndent(true);
  ui->scintilla->setTabIndents(true);
  ui->scintilla->setIndentationsUseTabs(false);
  ui->scintilla->setIndentationWidth(2);
  ui->scintilla->setEdgeColumn(80);
  ui->scintilla->setEdgeMode(QsciScintilla::EdgeLine);

  ui->scintilla->setMarginsFont(m_font);
  ui->scintilla->setMarginLineNumbers(0, true);
  ui->scintilla->setMarginWidth(0, "00000");

  ui->scintilla->setMarginWidth(1, 30);
  ui->scintilla->setMarginSensitivity(1, true);

  m_scintillaMarkerBreakpoint = ui->scintilla->markerDefine(QsciScintilla::Circle);
  ui->scintilla->setMarkerForegroundColor(Qt::black, m_scintillaMarkerBreakpoint);
  ui->scintilla->setMarkerBackgroundColor(Qt::white, m_scintillaMarkerBreakpoint);

  m_scintillaMarkerHalted = ui->scintilla->markerDefine(QsciScintilla::RightArrow);
  ui->scintilla->setMarkerForegroundColor(Qt::black, m_scintillaMarkerHalted);
  ui->scintilla->setMarkerBackgroundColor(Qt::green, m_scintillaMarkerHalted);

  ui->scintilla->connect(ui->scintilla, SIGNAL(marginClicked(int, int, Qt::KeyboardModifiers)), this, SLOT(slotMarginClicked(int, int, Qt::KeyboardModifiers)));

  ui->scintilla->setBraceMatching(QsciScintilla::SloppyBraceMatch);
  ui->scintilla->setMatchedBraceBackgroundColor(QColor("#ffe4e4"));

  ui->scintilla->setCaretLineVisible(true);
  ui->scintilla->setCaretLineBackgroundColor(QColor("#f1f1f1"));

  connect(ui->scintilla, SIGNAL(modificationChanged(bool)), this, SLOT(modificationChanged(bool)));

  m_curveColorList.push_back(Qt::blue);
  m_curveColorList.push_back(Qt::red);
  m_curveColorList.push_back(Qt::green);
  m_curveColorList.push_back(Qt::darkYellow);
  m_curveColorList.push_back(Qt::cyan);
  m_curveColorList.push_back(Qt::magenta);
  m_curveColorList.push_back(Qt::darkRed);
  m_curveColorList.push_back(Qt::darkGreen);
  m_curveColorList.push_back(Qt::darkBlue);
  m_curveColorList.push_back(Qt::darkCyan);
  m_curveColorList.push_back(Qt::darkMagenta);
  m_curveColorList.push_back(Qt::gray);
  m_curveColorList.push_back(Qt::black);

  ui->splitterDebugInfo->setSizes(m_debugInfoWidthDebugOff);

  m_globalVarFilter["string"] = true;
  m_globalVarFilter["package"] = true;
  m_globalVarFilter["os"] = true;
  m_globalVarFilter["dia"] = true;
  m_globalVarFilter["io"] = true;
  m_globalVarFilter["array"] = true;
  m_globalVarFilter["socket"] = true;
  m_globalVarFilter["coroutine"] = true;
  m_globalVarFilter["string_ex"] = true;
  m_globalVarFilter["crc"] = true;
  m_globalVarFilter["math_ex"] = true;
  m_globalVarFilter["gfx"] = true;
  m_globalVarFilter["vgfx"] = true;
  m_globalVarFilter["tab"] = true;
  m_globalVarFilter["bit"] = true;
  m_globalVarFilter["sys"] = true;
  m_globalVarFilter["math"] = true;
  m_globalVarFilter["debug"] = true;
  m_globalVarFilter["table"] = true;
  m_globalVarFilter["_G"] = true;
  m_globalVarFilter["_VERSION"] = true;

  ui->tableWidgetCallStack->setHorizontalHeaderItem(0, new QTableWidgetItem("function"));
  ui->tableWidgetCallStack->setHorizontalHeaderItem(1, new QTableWidgetItem("line"));

  ui->treeWidgetVariables->setHeaderLabels(QStringList() << "variable" << "value" << "type");

  m_globalVarsItem = 0;
  m_localVarsItem = 0;

  connect(ui->treeWidgetVariables, SIGNAL(itemExpanded(QTreeWidgetItem *)), this, SLOT(slotDebugInfoVariablesItemExpanded(QTreeWidgetItem *)));

  m_findDialog = new FindDialog(this);
  m_findDialog->setModal(false);
  m_findDialog->setTextEdit(ui->scintilla);

  m_findReplaceDialog = new FindReplaceDialog(this);
  m_findReplaceDialog->setModal(false);
  m_findReplaceDialog->setTextEdit(ui->scintilla);

  connect(ui->actionFind_next, SIGNAL(triggered()), m_findDialog, SLOT(findNext()));
  connect(ui->actionFind_previous, SIGNAL(triggered()), m_findDialog, SLOT(findPrev()));

  m_graphicsSceneColor = Qt::black;
  m_graphicsSceneFillColor = Qt::black;
  ui->graphicsViewVector->setScene(&m_graphicsScene);

  ui->gfxView->setRenderHints(QPainter::Antialiasing | QPainter::SmoothPixmapTransform);
  ui->graphicsViewVector->setRenderHints(QPainter::Antialiasing | QPainter::SmoothPixmapTransform);

  ui->console->setEnabled(true);

  ui->actionIndent->setText("Indent block\tTab");
  ui->actionUnindent->setText("Indent block\tShift+Tab");
  ui->actionUndo->setText("Undo\tCtrl+Z");
  ui->actionRedo->setText("Redo\tCtrl+Shift+Z");

  QStringList styles = QStyleFactory::keys();
  int         idx;

  ui->menuGUI_style->clear();
  m_actionGroupGuiStyle.setExclusive(true);

  for (idx = 0; idx < styles.size(); idx++)
  {
    QAction *action = ui->menuGUI_style->addAction(styles.at(idx));
    action->setCheckable(true);
    connect(action, SIGNAL(triggered()), this, SLOT(slotGuiStyle()));
    m_actionGroupGuiStyle.addAction(action);
  }
}

//-----------------------------------------------------------------------------

WindowMain::~WindowMain()
{
  cout.rdbuf(m_oldStreamBufOut);
  cerr.rdbuf(m_oldStreamBufErr);
  delete m_consoleStreamBufOut;
  delete m_consoleStreamBufErr;

  delete ui;
}

//-----------------------------------------------------------------------------

void WindowMain::SlotTableClear()
{
  ui->tableWidget->clear();
}

//-----------------------------------------------------------------------------

void WindowMain::SlotTableSetSize(int rows, int columns)
{
  ui->tableWidget->setRowCount(rows);
  ui->tableWidget->setColumnCount(columns);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotTableSetValue(int row, int column, const QString & value)
{
  ui->tableWidget->setItem(row - 1, column - 1, new QTableWidgetItem(value));
}

//-----------------------------------------------------------------------------

void WindowMain::SlotTableSetValuesInRow(int row, const QList<QString> &valueList)
{
  int numCols = ui->tableWidget->columnCount();
  int col;

  for (col = 0; (col < numCols) && (col < (int)valueList.size()); col++)
    ui->tableWidget->setItem(row - 1, col, new QTableWidgetItem(valueList[col]));
}

//-----------------------------------------------------------------------------

void WindowMain::SlotTableSetValuesInColumn(int column, const QList<QString> &valueList)
{
  int numRows = ui->tableWidget->rowCount();
  int row;

  for (row = 0; (row < numRows) && (row < (int)valueList.size()); row++)
    ui->tableWidget->setItem(row, column - 1, new QTableWidgetItem(valueList[row]));
}

//-----------------------------------------------------------------------------

void WindowMain::SlotTableSaveCsv(const QString &fileName)
{
  SaveTableDataAsCsv(fileName, ui->tableWidget);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotDiaDiagramNew(const QString & diagramName)
{
  if (m_diagramList.contains(diagramName))
    return;

  Diagram dia;

  dia.m_plot = new QCustomPlot(ui->tabWidgetDiagrams);
  ui->tabWidgetDiagrams->addTab(dia.m_plot, diagramName);
  dia.m_curveList.clear();
  dia.m_curveColorIndex = 0;

  m_diagramList[diagramName] = dia;

  dia.m_plot->setTitle(diagramName);
  dia.m_plot->setInteractions(QCustomPlot::iRangeDrag | QCustomPlot::iRangeZoom | QCustomPlot::iSelectAxes);
  dia.m_plot->setRangeDrag(Qt::Horizontal|Qt::Vertical);
  dia.m_plot->setRangeZoom(Qt::Horizontal|Qt::Vertical);
  dia.m_plot->legend->setVisible(true);
  dia.m_plot->replot();
  dia.m_plot->setContextMenuPolicy(Qt::CustomContextMenu);
  connect(dia.m_plot, SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(slotContextMenuRequest(QPoint)));
}

//-----------------------------------------------------------------------------

void WindowMain::SlotDiaDiagramDelete(const QString & diagramName)
{
  DiagramList::iterator diaIt = m_diagramList.find(diagramName);

  if (diaIt != m_diagramList.end())
  {
    Diagram dia = diaIt.value();
    m_diagramList.remove(diaIt.key());

    delete dia.m_plot;
  }
}

//-----------------------------------------------------------------------------

void WindowMain::SlotDiaDiagramDeleteAll()
{
  DiagramList::iterator diaIt;

  for (diaIt = m_diagramList.begin(); diaIt != m_diagramList.end(); ++diaIt)
  {
    delete diaIt.value().m_plot;
  }

  m_diagramList.clear();
}

//-----------------------------------------------------------------------------

void WindowMain::SlotDiaCurveNew(const QString & diagramName, const QString & curveName, int yAxis)
{
  DiagramList::iterator diaIt = m_diagramList.find(diagramName);
  if (diaIt == m_diagramList.end())
    return;

  Diagram & dia = diaIt.value();

  CurveList::iterator curveIt;
  curveIt = dia.m_curveList.find(curveName);
  if (curveIt != dia.m_curveList.end())
    return;

  QCPAxis *valueAxis;

  if (yAxis == 2)
  {
    valueAxis = dia.m_plot->yAxis2;
    dia.m_plot->yAxis2->setVisible(true);
  }
  else
    valueAxis = dia.m_plot->yAxis;

  QCPGraph *graph = dia.m_plot->addGraph(dia.m_plot->xAxis, valueAxis);

  graph->setName(curveName);
  graph->setPen(QPen(m_curveColorList[dia.m_curveColorIndex]));

  dia.m_curveColorIndex++;
  if (dia.m_curveColorIndex >= m_curveColorList.size())
    dia.m_curveColorIndex = 0;

  dia.m_curveList[curveName] = graph;
  dia.m_plot->replot();
}

//-----------------------------------------------------------------------------

void WindowMain::SlotDiaCurveDeleteAll(const QString & diagramName)
{
  DiagramList::iterator diaIt = m_diagramList.find(diagramName);
  if (diaIt == m_diagramList.end())
    return;

  Diagram & dia = diaIt.value();

  dia.m_curveColorIndex = 0;
  dia.m_curveList.clear();
  dia.m_plot->clearGraphs();
  dia.m_plot->replot();
}

//-----------------------------------------------------------------------------

void WindowMain::SlotDiaCurveDelete(const QString & diagramName, const QString & curveName)
{
  DiagramList::iterator diaIt = m_diagramList.find(diagramName);
  if (diaIt == m_diagramList.end())
    return;

  Diagram & dia = diaIt.value();

  CurveList::iterator curveIt;

  curveIt = dia.m_curveList.find(curveName);
  if (curveIt != dia.m_curveList.end())
  {
    dia.m_plot->removeGraph(curveIt.value());
    dia.m_plot->rescaleAxes();
    dia.m_plot->replot();
    dia.m_curveList.remove(curveIt.key());
  }
}

//-----------------------------------------------------------------------------

void WindowMain::SlotDiaCurveClearData(const QString & diagramName, const QString &curveName)
{
  DiagramList::iterator diaIt = m_diagramList.find(diagramName);
  if (diaIt == m_diagramList.end())
    return;

  Diagram & dia = diaIt.value();

  CurveList::iterator curveIt;

  curveIt = dia.m_curveList.find(curveName);
  if (curveIt != dia.m_curveList.end())
  {
    curveIt.value()->clearData();
    dia.m_plot->replot();
  }
}

//-----------------------------------------------------------------------------

void WindowMain::SlotDiaCurveSetData(const QString & diagramName, const QString &curveName, const QVector <double> & x, const QVector <double> & y)
{
  DiagramList::iterator diaIt = m_diagramList.find(diagramName);
  if (diaIt == m_diagramList.end())
    return;

  Diagram & dia = diaIt.value();

  CurveList::iterator curveIt;

  curveIt = dia.m_curveList.find(curveName);
  if (curveIt != dia.m_curveList.end())
  {
    curveIt.value()->setData(x, y);
    dia.m_plot->rescaleAxes();
    dia.m_plot->replot();
  }
}

//-----------------------------------------------------------------------------

void WindowMain::SlotDiaCurveAppendData(const QString & diagramName, const QString &curveName, double x, double y)
{
  DiagramList::iterator diaIt = m_diagramList.find(diagramName);
  if (diaIt == m_diagramList.end())
    return;

  Diagram & dia = diaIt.value();

  CurveList::iterator curveIt;

  curveIt = dia.m_curveList.find(curveName);
  if (curveIt != dia.m_curveList.end())
  {
    curveIt.value()->addData(x, y);
    dia.m_plot->rescaleAxes();
    dia.m_plot->replot();
  }
}

//-----------------------------------------------------------------------------

void WindowMain::SetGuiState(WindowMain::GuiState guiState)
{
  switch (guiState)
  {
  case GS_RUNNING:
    ui->actionNew->setEnabled(false);
    ui->actionLoad->setEnabled(false);
    ui->actionSave->setEnabled(false);
    ui->actionSave_As->setEnabled(false);
    ui->actionRun->setEnabled(false);
    ui->actionStop->setEnabled(true);
    ui->actionDebug->setEnabled(false);
    ui->actionHalt->setEnabled(ui->actionDebug->isChecked());
    ui->actionContinue->setEnabled(false);
    ui->actionStepIn->setEnabled(false);
    ui->actionStepOver->setEnabled(false);
    ui->actionStepOut->setEnabled(false);
    ui->treeWidgetVariables->setEnabled(false);
    ui->tableWidgetCallStack->setEnabled(false);
    ui->scintilla->setReadOnly(true);
    ui->menuRecent_files->setEnabled(false);
    ui->actionReset_Lua->setEnabled(false);
    break;

  case GS_HALTED:
    ui->actionNew->setEnabled(false);
    ui->actionLoad->setEnabled(false);
    ui->actionSave->setEnabled(false);
    ui->actionSave_As->setEnabled(false);
    ui->actionRun->setEnabled(false);
    ui->actionStop->setEnabled(true);
    ui->actionDebug->setEnabled(false);
    ui->actionHalt->setEnabled(false);
    ui->actionContinue->setEnabled(true);
    ui->actionStepIn->setEnabled(true);
    ui->actionStepOver->setEnabled(true);
    ui->actionStepOut->setEnabled(true);
    ui->treeWidgetVariables->setEnabled(true);
    ui->tableWidgetCallStack->setEnabled(true);
    ui->scintilla->setReadOnly(true);
    ui->menuRecent_files->setEnabled(false);
    ui->actionReset_Lua->setEnabled(false);
    break;

  case GS_STOPPED:
  default:
    ui->actionNew->setEnabled(true);
    ui->actionLoad->setEnabled(true);
    ui->actionSave->setEnabled(true);
    ui->actionSave_As->setEnabled(true);
    ui->actionRun->setEnabled(true);
    ui->actionStop->setEnabled(false);
    ui->actionDebug->setEnabled(true);
    ui->actionHalt->setEnabled(false);
    ui->actionContinue->setEnabled(false);
    ui->actionStepIn->setEnabled(false);
    ui->actionStepOver->setEnabled(false);
    ui->actionStepOut->setEnabled(false);
    ui->treeWidgetVariables->setEnabled(false);
    ui->tableWidgetCallStack->setEnabled(false);
    ui->scintilla->setReadOnly(false);
    ui->menuRecent_files->setEnabled(true);
    ui->actionReset_Lua->setEnabled(true);
    break;
  }
}

//-----------------------------------------------------------------------------

void WindowMain::LuaStarted()
{
  SetGuiState(GS_RUNNING);
  ui->scintilla->markerDeleteAll(m_scintillaMarkerHalted);
}

//-----------------------------------------------------------------------------

void WindowMain::LuaStopped()
{
  SetGuiState(GS_STOPPED);
  ui->scintilla->markerDeleteAll(m_scintillaMarkerHalted);
}

//-----------------------------------------------------------------------------

void WindowMain::LuaHalted()
{
  SetGuiState(GS_HALTED);

  int line = m_luaThread->CurrentLine();

  if (line > 0)
  {
    ui->scintilla->setCursorPosition(line - 1, 0);
    ui->scintilla->markerAdd(line - 1, m_scintillaMarkerHalted);
  }

  ui->scintilla->ensureCursorVisible();
  ui->scintilla->update();

  ShowDebugInfo();
}

//-----------------------------------------------------------------------------

void WindowMain::LuaContinued()
{
  SetGuiState(GS_RUNNING);
  ui->scintilla->markerDeleteAll(m_scintillaMarkerHalted);
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionStop_triggered()
{
  g_lua_stop = 1;
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionRun_triggered()
{
  if (m_luaThread->IsRunning())
    return;

  ui->console->Write("\n");
  m_luaThread->Execute(ui->scintilla->text());
  UpdateBreakpointsInLuaDebugger();
}

//-----------------------------------------------------------------------------

void WindowMain::AddToRecentFiles(const QString & fileName)
{
  int idx;

  idx = m_recentFiles.indexOf(fileName);
  while (idx != -1)
  {
    m_recentFiles.removeAt(idx);
    idx = m_recentFiles.indexOf(fileName);
  }

  while (m_recentFiles.count() >= MAX_RECENT_FILES)
    m_recentFiles.removeFirst();

  m_recentFiles.append(fileName);

  UpdateRecentFilesMenu();
}

//-----------------------------------------------------------------------------

void WindowMain::UpdateRecentFilesMenu()
{
  ui->menuRecent_files->clear();

  int idx;

  for (idx = m_recentFiles.count() - 1; idx >= 0; idx--)
  {
    QAction *action = ui->menuRecent_files->addAction(m_recentFiles.at(idx));
    connect(action, SIGNAL(triggered()), this, SLOT(slotOpenRecentFile()));
  }
}

//-----------------------------------------------------------------------------

void WindowMain::slotOpenRecentFile()
{
  if (!ContinueActionIfModified())
    return;

  QAction *action = qobject_cast<QAction *>(sender());
  if (action)
  {
    QString fileName = action->text();
    if (!fileName.isEmpty())
      LoadFile(fileName);
  }
}

//-----------------------------------------------------------------------------

void WindowMain::slotGuiStyle()
{
  QAction *action = qobject_cast<QAction *>(sender());
  if (action)
  {
    action->setChecked(true);
    QString style = action->text();
    if (!style.isEmpty())
      QApplication::setStyle(style);
  }
}

//-----------------------------------------------------------------------------

bool WindowMain::LoadFile(const QString &fileName)
{
  bool ret = false;

  if (!fileName.isEmpty())
  {
    try
    {
      QFile file(fileName);

      file.open(QIODevice::ReadOnly | QIODevice::Text);
      ui->scintilla->setText(file.readAll());

      file.close();

      m_luaThread->ClearBreakpoints();
      ui->scintilla->setModified(false);
      m_currentFile = fileName;
      AddToRecentFiles(fileName);
      ret = true;
    }
    catch (exception & exc)
    {
      QMessageBox::critical(this, "ERROR", QString::fromLocal8Bit(exc.what()));
    }
  }

  UpdateWindowTitle();

  return ret;
}

//-----------------------------------------------------------------------------

void WindowMain::SaveFile(QString fileName)
{
  if (!fileName.isEmpty())
  {
    try
    {
      QFile file(fileName);

      file.open(QIODevice::WriteOnly | QIODevice::Text);
      if (file.write(ui->scintilla->text().toAscii()) == -1)
        throw runtime_error("can not save file");

      file.close();

      ui->scintilla->setModified(false);
      m_currentFile = fileName;
      AddToRecentFiles(fileName);
    }
    catch (exception & exc)
    {
      QMessageBox::critical(this, "ERROR", QString::fromLocal8Bit(exc.what()));
    }
  }

  UpdateWindowTitle();
}

//-----------------------------------------------------------------------------

void WindowMain::UpdateWindowTitle()
{
  QString title = "Lua-Quick-Try-Out";

  if (!m_currentFile.isEmpty())
    title += "   " + m_currentFile;

  if (ui->scintilla->isModified())
    title += "   (modified)";

  setWindowTitle(title);
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionLoad_triggered()
{
  if (!ContinueActionIfModified())
    return;

  QFileDialog dlg(this, "Load file", m_currentFile, "Lua script (*.lua)");

  dlg.setFileMode(QFileDialog::ExistingFile);
  dlg.setDefaultSuffix("lua");
  dlg.setViewMode(QFileDialog::Detail);

  if (dlg.exec())
  {
    QString fileName = dlg.selectedFiles()[0];
    LoadFile(fileName);
  }
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionSave_As_triggered()
{
  QFileDialog dlg(this, "Save file", m_currentFile, "Lua script (*.lua)");

  dlg.setFileMode(QFileDialog::AnyFile);
  dlg.setDefaultSuffix("lua");
  dlg.setViewMode(QFileDialog::Detail);
  dlg.setConfirmOverwrite(true);
  dlg.setAcceptMode(QFileDialog::AcceptSave);

  if (dlg.exec())
  {
    QString fileName = dlg.selectedFiles()[0];
    SaveFile(fileName);
  }
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionSave_triggered()
{
  if (m_currentFile.isEmpty())
    on_actionSave_As_triggered();
  else
    SaveFile(m_currentFile);
}

//-----------------------------------------------------------------------------

void WindowMain::ProgramStarted()
{
  try
  {
    QSettings appSettings("BriSoft", "LuaQuickTryOut");
    int       idx;

    m_findReplaceDialog->readSettings(appSettings);

    appSettings.beginGroup("WindowMain");
    resize(appSettings.value("size", size()).toSize());
    move(appSettings.value("pos", pos()).toPoint());
    ui->splitter->restoreState(appSettings.value("splitterSizes").toByteArray());

    QString style = appSettings.value("Style", QString()).toString();
    QApplication::setStyle(style);
    style = QApplication::style()->objectName();

    QList <QAction *> actions = m_actionGroupGuiStyle.actions();

    for (idx = 0; idx < actions.size(); idx++)
    {
      if (actions.at(idx)->text().toLower() == style.toLower())
      {
        actions.at(idx)->setChecked(true);
        break;
      }
    }

    m_debugInfoWidthDebugOn.clear();
    m_debugInfoWidthDebugOn.push_back(700);
    m_debugInfoWidthDebugOn.push_back(300);
    m_debugInfoWidthDebugOn[0] = appSettings.value("DebugInfoWidthDebugOn0", 700).toInt();
    m_debugInfoWidthDebugOn[1] = appSettings.value("DebugInfoWidthDebugOn1", 300).toInt();

    m_debugInfoWidthDebugOff.clear();
    m_debugInfoWidthDebugOff.push_back(1000);
    m_debugInfoWidthDebugOff.push_back(0);
    m_debugInfoWidthDebugOff[0] = appSettings.value("DebugInfoWidthDebugOff0", 1000).toInt();
    m_debugInfoWidthDebugOff[1] = appSettings.value("DebugInfoWidthDebugOff1", 0).toInt();

    appSettings.endGroup();

    ui->splitterDebugInfo->setSizes(m_debugInfoWidthDebugOff);

    appSettings.beginGroup("CurrentFile");
    m_currentFile = appSettings.value("LuaFile", "").toString();
    appSettings.endGroup();

    if (!m_currentFile.isEmpty())
    {
      if (!LoadFile(m_currentFile))
        m_currentFile.clear();
    }

    m_recentFiles.clear();

    appSettings.beginGroup("RecentFiles");

    for (idx = 0; idx < MAX_RECENT_FILES; idx++)
    {
      QString fileName = appSettings.value("RecentFile" + QString::number(idx), "").toString();
      if (!fileName.isEmpty())
        m_recentFiles.append(fileName);
    }

    appSettings.endGroup();

    UpdateRecentFilesMenu();

    ui->scintilla->setModified(false);
    UpdateWindowTitle();
  }
  catch (exception & exc)
  {
    QMessageBox::critical(this, "ERROR", QString::fromLocal8Bit(exc.what()));
  }
}

//-----------------------------------------------------------------------------

void WindowMain::closeEvent(QCloseEvent *event)
{
  if (!ContinueActionIfModified())
  {
    event->setAccepted(false);
    return;
  }

  if (m_luaThread->isRunning())
  {
    g_lua_stop = 1;
    m_luaThread->wait(5000);
  }

  if (ui->actionDebug->isChecked())
    m_debugInfoWidthDebugOn = ui->splitterDebugInfo->sizes();
  else
    m_debugInfoWidthDebugOff = ui->splitterDebugInfo->sizes();

  QSettings appSettings("BriSoft", "LuaQuickTryOut");

  m_findReplaceDialog->writeSettings(appSettings);

  appSettings.beginGroup("WindowMain");
  appSettings.setValue("size", size());
  appSettings.setValue("pos", pos());
  appSettings.setValue("splitterSizes", ui->splitter->saveState());
  appSettings.setValue("Style", QApplication::style()->objectName());

  if (m_debugInfoWidthDebugOn.size() == 2)
  {
    appSettings.setValue("DebugInfoWidthDebugOn0", m_debugInfoWidthDebugOn[0]);
    appSettings.setValue("DebugInfoWidthDebugOn1", m_debugInfoWidthDebugOn[1]);
  }
  if (m_debugInfoWidthDebugOff.size() == 2)
  {
    appSettings.setValue("DebugInfoWidthDebugOff0", m_debugInfoWidthDebugOff[0]);
    appSettings.setValue("DebugInfoWidthDebugOff1", m_debugInfoWidthDebugOff[1]);
  }

  appSettings.endGroup();

  appSettings.beginGroup("CurrentFile");
  appSettings.setValue("LuaFile", m_currentFile);
  appSettings.endGroup();

  int idx;

  appSettings.beginGroup("RecentFiles");

  for (idx = 0; idx < m_recentFiles.count(); idx++)
    appSettings.setValue("RecentFile" + QString::number(idx), m_recentFiles.at(idx));

  for (; idx < MAX_RECENT_FILES; idx++)
    appSettings.setValue("RecentFile" + QString::number(idx), "");

  appSettings.endGroup();
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionFunction_help_triggered()
{
  m_winFuncHelp.setVisible(true);
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionAbout_triggered()
{
  m_winAbout.setVisible(true);
}

//-----------------------------------------------------------------------------

void WindowMain::modificationChanged(bool changed)
{
  (void)changed;
  UpdateWindowTitle();
}

//-----------------------------------------------------------------------------

bool WindowMain::ContinueActionIfModified()
{
  if (ui->scintilla->isModified())
  {
    int ret = QMessageBox::warning(this, "Warning",
      "Source code was modified but not saved. Do you really want to continue and loose changes?",
      QMessageBox::Yes | QMessageBox::No);

    if (ret == QMessageBox::No)
    {
      return false;
    }
  }

  return true;
}

//-----------------------------------------------------------------------------

void WindowMain::slotContextMenuRequest(QPoint pos)
{
  QCustomPlot *plot = dynamic_cast <QCustomPlot *>(sender());
  m_currentDiagramContextMenu = plot;

  if (plot)
  {
    QMenu *menu = new QMenu(this);

    menu->setAttribute(Qt::WA_DeleteOnClose);
    menu->addAction("zoom to fit", this, SLOT(slotZoomToFit()));
    menu->addAction("save image", this, SLOT(slotDiagramSaveImage()));
    menu->addAction("save data", this, SLOT(slotDiagramSaveCsv()));

    menu->popup(plot->mapToGlobal(pos));
  }
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionSave_image_triggered()
{
  QFileDialog dlg(this, "Save graphic as image", "", "PNG (*.png)");

  dlg.setFileMode(QFileDialog::AnyFile);
  dlg.setDefaultSuffix("png");
  dlg.setViewMode(QFileDialog::Detail);
  dlg.setConfirmOverwrite(true);
  dlg.setAcceptMode(QFileDialog::AcceptSave);

  if (dlg.exec())
  {
    QString fileName = dlg.selectedFiles()[0];
    ui->gfxView->m_gfxItemPixmap->m_pixmap.save(fileName);
  }
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionSave_data_triggered()
{
  QFileDialog dlg(this, "Save table data", "", "CSV (*.csv)");

  dlg.setFileMode(QFileDialog::AnyFile);
  dlg.setDefaultSuffix("csv");
  dlg.setViewMode(QFileDialog::Detail);
  dlg.setConfirmOverwrite(true);
  dlg.setAcceptMode(QFileDialog::AcceptSave);

  if (dlg.exec())
  {
    QString fileName = dlg.selectedFiles()[0];
    SaveTableDataAsCsv(fileName, ui->tableWidget);
  }
}

//-----------------------------------------------------------------------------

void WindowMain::slotZoomToFit()
{
  QCustomPlot *plot = m_currentDiagramContextMenu;
  if (plot)
  {
    plot->rescaleAxes();
    plot->replot();
  }
}

//-----------------------------------------------------------------------------

void WindowMain::slotDiagramSaveImage()
{
  QCustomPlot *plot = m_currentDiagramContextMenu;
  if (plot)
  {
    QFileDialog dlg(this, "Save diagram as image", "", "PNG (*.png)");

    dlg.setFileMode(QFileDialog::AnyFile);
    dlg.setDefaultSuffix("png");
    dlg.setViewMode(QFileDialog::Detail);
    dlg.setConfirmOverwrite(true);
    dlg.setAcceptMode(QFileDialog::AcceptSave);

    if (dlg.exec())
    {
      QString fileName = dlg.selectedFiles()[0];
      plot->savePng(fileName);
    }
  }
}

//-----------------------------------------------------------------------------

void WindowMain::slotDiagramSaveCsv()
{
  QCustomPlot *plot = m_currentDiagramContextMenu;
  if (plot)
  {
    QFileDialog dlg(this, "Save diagram data", "", "CSV (*.csv)");

    dlg.setFileMode(QFileDialog::AnyFile);
    dlg.setDefaultSuffix("csv");
    dlg.setViewMode(QFileDialog::Detail);
    dlg.setConfirmOverwrite(true);
    dlg.setAcceptMode(QFileDialog::AcceptSave);

    if (dlg.exec())
    {
      QString fileName = dlg.selectedFiles()[0];
      SaveDiagramDataAsCsv(fileName, plot);
    }
  }
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionDebug_triggered()
{
  bool chk = ui->actionDebug->isChecked();

  try
  {
    m_luaThread->Debug(chk);
  }
  catch (exception & exc)
  {
    ui->actionDebug->setChecked(!chk);
    cout << "ERROR: " << exc.what() << endl;
  }

  if (ui->actionDebug->isChecked())
  {
    ui->actionDebug->setText("Debug On");
    m_debugInfoWidthDebugOff = ui->splitterDebugInfo->sizes();
    ui->splitterDebugInfo->setSizes(m_debugInfoWidthDebugOn);
    ui->scintilla->setCaretLineBackgroundColor(QColor("#FFA0A0"));
    ui->scintilla->setMarkerBackgroundColor(Qt::red, m_scintillaMarkerBreakpoint);
    UpdateBreakpointsInLuaDebugger();
  }
  else
  {
    ui->actionDebug->setText("Debug Off");
    m_debugInfoWidthDebugOn = ui->splitterDebugInfo->sizes();
    ui->splitterDebugInfo->setSizes(m_debugInfoWidthDebugOff);
    ui->scintilla->setCaretLineBackgroundColor(QColor("#f1f1f1"));
    ui->scintilla->setMarkerBackgroundColor(Qt::white, m_scintillaMarkerBreakpoint);
  }

  update();
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionHalt_triggered()
{
  try
  {
    m_luaThread->DebuggerHalt();
  }
  catch (exception & exc)
  {
    cout << "ERROR: " << exc.what() << endl;
  }
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionContinue_triggered()
{
  try
  {
    m_luaThread->DebuggerContinue();
  }
  catch (exception & exc)
  {
    cout << "ERROR: " << exc.what() << endl;
  }
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionClear_all_breakpoints_triggered()
{
  try
  {
    m_luaThread->ClearBreakpoints();
    ui->scintilla->markerDeleteAll(m_scintillaMarkerBreakpoint);
  }
  catch (exception & exc)
  {
    cout << "ERROR: " << exc.what() << endl;
  }

  ui->scintilla->update();
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionStepIn_triggered()
{
  try
  {
    m_luaThread->DebuggerStepIn();
  }
  catch (exception & exc)
  {
    cout << "ERROR: " << exc.what() << endl;
  }
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionStepOut_triggered()
{
  try
  {
    m_luaThread->DebuggerStepOut();
  }
  catch (exception & exc)
  {
    cout << "ERROR: " << exc.what() << endl;
  }
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionStepOver_triggered()
{
  try
  {
    m_luaThread->DebuggerStepOver();
  }
  catch (exception & exc)
  {
    cout << "ERROR: " << exc.what() << endl;
  }
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionNew_triggered()
{
  if (!ContinueActionIfModified())
    return;

  ui->scintilla->clear();
  ui->scintilla->setModified(false);
  m_luaThread->ClearBreakpoints();
  m_currentFile.clear();
  UpdateWindowTitle();
}

//-----------------------------------------------------------------------------

void WindowMain::ShowDebugInfo()
{
  ShowDebugInfoCallStack();
  ShowDebugInfoVariables();
}

//-----------------------------------------------------------------------------

void WindowMain::ShowDebugInfoCallStack()
{
  try
  {
    CallStack callStack = m_luaThread->DebuggerCallStack();
    int idx;

    ui->tableWidgetCallStack->clear();
    ui->tableWidgetCallStack->setRowCount(callStack.m_callStack.size());
    ui->tableWidgetCallStack->setHorizontalHeaderItem(0, new QTableWidgetItem("function"));
    ui->tableWidgetCallStack->setHorizontalHeaderItem(1, new QTableWidgetItem("line"));

    for (idx = 0; idx < callStack.m_callStack.size(); idx++)
    {
      StackFrame & frame = callStack.m_callStack[idx];

      ui->tableWidgetCallStack->setItem(idx, 0, new QTableWidgetItem(frame.m_functionName));
      ui->tableWidgetCallStack->setItem(idx, 1, new QTableWidgetItem(QString::number(frame.m_currentLine)));
    }
  }
  catch (exception & exc)
  {
    cout << "ERROR: " << exc.what() << endl;
  }
}

//-----------------------------------------------------------------------------

void WindowMain::ShowDebugInfoVariables()
{
  try
  {
    ui->treeWidgetVariables->clear();

    VariableList localVars = m_luaThread->DebuggerGetLocalVariables();
    VariableList globalVars = m_luaThread->DebuggerGetGlobalVariables();
    int idx;

    m_globalVarsItem = new QTreeWidgetItem(QStringList("global"));
    m_localVarsItem = new QTreeWidgetItem(QStringList("local"));

    for (idx = 0; idx < globalVars.m_varList.size(); idx++)
    {
      Variable & var = globalVars.m_varList[idx];

      if (!m_globalVarFilter.contains(var.m_name))
      {
        QStringList stringList(var.m_name);
        stringList.append(var.m_value);
        stringList.append(var.m_type);
        QTreeWidgetItem *item = new QTreeWidgetItem(stringList);

        if (var.m_type == "table")
        {
          item->addChild(new QTreeWidgetItem(QStringList("...")));
          item->setData(0, Qt::UserRole, var.m_key);
        }

        m_globalVarsItem->addChild(item);
      }
    }

    for (idx = 0; idx < localVars.m_varList.size(); idx++)
    {
      Variable & var = localVars.m_varList[idx];

      QStringList stringList(var.m_name);
      stringList.append(var.m_value);
      stringList.append(var.m_type);
      QTreeWidgetItem *item = new QTreeWidgetItem(stringList);

      if (var.m_type == "table")
      {
        item->addChild(new QTreeWidgetItem(QStringList("...")));
        item->setData(0, Qt::UserRole, var.m_key);
      }

      m_localVarsItem->addChild(item);
    }

    ui->treeWidgetVariables->addTopLevelItem(m_localVarsItem);
    ui->treeWidgetVariables->addTopLevelItem(m_globalVarsItem);
    m_localVarsItem->setExpanded(true);
    m_globalVarsItem->setExpanded(true);
  }
  catch (exception & exc)
  {
    cout << "ERROR: " << exc.what() << endl;
  }
}

//-----------------------------------------------------------------------------

void WindowMain::slotDebugInfoVariablesItemExpanded(QTreeWidgetItem *item)
{
  if ((item == m_globalVarsItem) || (item == m_localVarsItem))
    return;

  if (item->childCount() == 1)
  {
    QTreeWidgetItem *child = item->child(0);
    if ((child->text(0) == "...") && (child->text(1).length() == 0) && (child->text(2).length() == 0))
    {
      QVariantList    path;
      QTreeWidgetItem *pathItem = item;
      VariableList    varList;

      while (pathItem)
      {
        if (pathItem == m_localVarsItem)
        {
          varList = m_luaThread->DebuggerGetLocalVariablesFromPath(path);
          break;
        }
        else if (pathItem == m_globalVarsItem)
        {
          varList = m_luaThread->DebuggerGetGlobalVariablesFromPath(path);
          break;
        }

        path.push_front(pathItem->data(0, Qt::UserRole));
        pathItem = pathItem->parent();
      }

      qDeleteAll(item->takeChildren());

      int idx;

      for (idx = 0; (idx < varList.m_varList.size()) && (idx <= DEBUG_INFO_MAX_ARRAY_ITEMS); idx++)
      {
        Variable & var = varList.m_varList[idx];

        QStringList stringList(var.m_name);
        stringList.append(var.m_value);
        stringList.append(var.m_type);
        child = new QTreeWidgetItem(stringList);

        if (var.m_type == "table")
        {
          child->addChild(new QTreeWidgetItem(QStringList("...")));
          child->setData(0, Qt::UserRole, var.m_key);
        }

        item->addChild(child);
      }
    }
  }
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionReset_Lua_triggered()
{
  try
  {
    ui->actionDebug->setChecked(false);
    ui->actionDebug->setText("Debug Off");
    m_debugInfoWidthDebugOn = ui->splitterDebugInfo->sizes();
    ui->splitterDebugInfo->setSizes(m_debugInfoWidthDebugOff);
    update();

    m_luaThread->ResetLua();
  }
  catch (exception & exc)
  {
    cout << "ERROR: " << exc.what() << endl;
  }
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionFind_and_replace_triggered()
{
  m_findReplaceDialog->show();
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionFind_triggered()
{
  m_findDialog->show();
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionIndent_triggered()
{
  QKeyEvent keyEvent(QKeyEvent::KeyPress, Qt::Key_Tab, Qt::NoModifier);
  QApplication::sendEvent(ui->scintilla, &keyEvent);
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionUnindent_triggered()
{
  QKeyEvent keyEvent(QKeyEvent::KeyPress, Qt::Key_Backtab, Qt::NoModifier);
  QApplication::sendEvent(ui->scintilla, &keyEvent);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxClear()
{
  m_graphicsSceneColor = Qt::black;
  m_graphicsSceneFillColor = Qt::black;
  m_graphicsScene.clear();

  //m_graphicsScene.update();
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxSetColor(int red, int green, int blue)
{
  m_graphicsSceneColor = QColor(red, green, blue);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxSetColor(const QString & name)
{
  m_graphicsSceneColor = QColor(name);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxSetFillColor(int red, int green, int blue)
{
  m_graphicsSceneFillColor = QColor(red, green, blue);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxSetFillColor(const QString & name)
{
  m_graphicsSceneFillColor = QColor(name);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxDrawLine(qreal x1, qreal y1, qreal x2, qreal y2)
{
  m_graphicsScene.addLine(x1, y1, x2, y2, m_graphicsSceneColor);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxDrawCircle(qreal x, qreal y, qreal r)
{
  m_graphicsScene.addEllipse(x - r, y - r, 2 * r, 2 * r, m_graphicsSceneColor);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxDrawEllipse(qreal x, qreal y, qreal rx, qreal ry)
{
  m_graphicsScene.addEllipse(x - rx, y - ry, 2 * rx, 2 * ry, m_graphicsSceneColor);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxDrawRect(qreal x, qreal y, qreal w, qreal h)
{
  m_graphicsScene.addRect(x, y, w, h, m_graphicsSceneColor);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxDrawPolygon(const QPolygonF &polygon)
{
  m_graphicsScene.addPolygon(polygon, m_graphicsSceneColor);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxFillCircle(qreal x, qreal y, qreal r)
{
  m_graphicsScene.addEllipse(x - r, y - r, 2 * r, 2 * r, m_graphicsSceneColor, m_graphicsSceneFillColor);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxFillEllipse(qreal x, qreal y, qreal rx, qreal ry)
{
  m_graphicsScene.addEllipse(x - rx, y - ry, 2 * rx, 2 * ry, m_graphicsSceneColor, m_graphicsSceneFillColor);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxFillRect(qreal x, qreal y, qreal w, qreal h)
{
  m_graphicsScene.addRect(x, y, w, h, m_graphicsSceneColor, m_graphicsSceneFillColor);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxFillPolygon(const QPolygonF &polygon)
{
  m_graphicsScene.addPolygon(polygon, m_graphicsSceneColor, m_graphicsSceneFillColor);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxDrawText(qreal x, qreal y, const QString &text)
{
  QGraphicsTextItem *item = m_graphicsScene.addText(text);
  item->moveBy(x, y);
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionSave_image_2_triggered()
{
  QFileDialog dlg(this, "Save vector graphic as image", "", "PNG (*.png)");

  dlg.setFileMode(QFileDialog::AnyFile);
  dlg.setDefaultSuffix("png");
  dlg.setViewMode(QFileDialog::Detail);
  dlg.setConfirmOverwrite(true);
  dlg.setAcceptMode(QFileDialog::AcceptSave);

  if (dlg.exec())
  {
    QString fileName = dlg.selectedFiles()[0];

    QPixmap pixmap(1280, 1024);
    QPainter painter(&pixmap);

    painter.eraseRect(pixmap.rect());
    m_graphicsScene.render(&painter);
    pixmap.save(fileName);
  }
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionSave_SVG_triggered()
{
  QFileDialog dlg(this, "Save vector graphic as SVG", "", "SVG (*.svg)");

  dlg.setFileMode(QFileDialog::AnyFile);
  dlg.setDefaultSuffix("svg");
  dlg.setViewMode(QFileDialog::Detail);
  dlg.setConfirmOverwrite(true);
  dlg.setAcceptMode(QFileDialog::AcceptSave);

  if (dlg.exec())
  {
    QString       fileName = dlg.selectedFiles()[0];
    QSvgGenerator svgGen;
    QRectF        sceneRect = m_graphicsScene.sceneRect();

    svgGen.setSize(QSize(sceneRect.width(), sceneRect.height()));
    svgGen.setViewBox(QRectF(0, 0, sceneRect.width(), sceneRect.height()));

    svgGen.setFileName(fileName);

    QPainter painter(&svgGen);
    painter.eraseRect(QRectF(0, 0, sceneRect.width(), sceneRect.height()));
    m_graphicsScene.render(&painter);
  }
}

//-----------------------------------------------------------------------------

void WindowMain::SlotDiaDiagramSaveImg(const QString &diagramName, const QString &fileName)
{
  DiagramList::iterator diaIt = m_diagramList.find(diagramName);

  if (diaIt != m_diagramList.end())
  {
    diaIt.value().m_plot->savePng(fileName);
  }
}

//-----------------------------------------------------------------------------

void WindowMain::SlotDiaDiagramSaveCsv(const QString &diagramName, const QString &fileName)
{
  DiagramList::iterator diaIt = m_diagramList.find(diagramName);

  if (diaIt != m_diagramList.end())
  {
    SaveDiagramDataAsCsv(fileName, diaIt.value().m_plot);
  }
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxSaveImg(const QString &fileName, int width, int height)
{
  QPixmap pixmap(width, height);
  QPainter painter(&pixmap);

  painter.eraseRect(pixmap.rect());
  m_graphicsScene.render(&painter);
  pixmap.save(fileName);
}

//-----------------------------------------------------------------------------

void WindowMain::SlotVecGfxSaveCsv(const QString &fileName)
{
  QSvgGenerator svgGen;
  QRectF        sceneRect = m_graphicsScene.sceneRect();

  svgGen.setSize(QSize(sceneRect.width(), sceneRect.height()));
  svgGen.setViewBox(QRectF(0, 0, sceneRect.width(), sceneRect.height()));

  svgGen.setFileName(fileName);

  QPainter painter(&svgGen);
  painter.eraseRect(QRectF(0, 0, sceneRect.width(), sceneRect.height()));
  m_graphicsScene.render(&painter);
}

//-----------------------------------------------------------------------------

void WindowMain::SaveDiagramDataAsCsv(const QString &fileName, QCustomPlot *plot)
{
  QFile   file(fileName);

  file.open(QIODevice::WriteOnly);

  QTextStream stream(&file);
  int   graphIdx;
  QList <QList <QCPData> > data;

  for (graphIdx = 0; graphIdx < plot->graphCount(); graphIdx++)
  {
    QCPGraph *graph = plot->graph(graphIdx);

    if (graphIdx > 0)
      stream << "\t\t";

    stream << graph->name();

    data.push_back(graph->data()->values());
  }
  stream << "\n";

  int   idx, dataIdx;
  bool  notDone = true;

  for (idx = 0; notDone; idx++)
  {
    notDone = false;

    QString line;

    for (dataIdx = 0; dataIdx < data.size(); dataIdx++)
    {
      if (dataIdx > 0)
        line += "\t";

      if (idx < data[dataIdx].count())
      {
        notDone = true;
        line += QString::number(data[dataIdx][idx].key);
        line += "\t";
        line += QString::number(data[dataIdx][idx].value);
      }
    }

    if (notDone)
      stream << line << "\n";
  }

  file.close();
}

//-----------------------------------------------------------------------------

void WindowMain::SaveTableDataAsCsv(const QString &fileName, QTableWidget *table)
{
  QFile   file(fileName);

  file.open(QIODevice::WriteOnly);

  QTextStream stream(&file);

  int numRows = table->rowCount();
  int numCols = table->columnCount();
  int row, col;


  for (row = 0; row < numRows; row++)
  {
    for (col = 0; col < numCols; col++)
    {
      if (col > 0)
        stream << "\t";

      QTableWidgetItem *item = table->item(row, col);

      if (item)
      {
        stream << item->data(Qt::DisplayRole).toString();
      }
    }

    stream << "\n";
  }

  file.close();
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionUndo_triggered()
{
  ui->scintilla->undo();
}

//-----------------------------------------------------------------------------

void WindowMain::on_actionRedo_triggered()
{
  ui->scintilla->redo();
}

//-----------------------------------------------------------------------------

void WindowMain::slotMarginClicked(int margin, int line, Qt::KeyboardModifiers state)
{
  (void)state;

  if (margin == 1)
  {
    if (ui->scintilla->markersAtLine(line) & (1 << m_scintillaMarkerBreakpoint))
    {
      ui->scintilla->markerDelete(line, m_scintillaMarkerBreakpoint);
      m_luaThread->ClearBreakpoint(line + 1);
    }
    else
    {
      ui->scintilla->markerAdd(line, m_scintillaMarkerBreakpoint);
      m_luaThread->SetBreakpoint(line + 1);
    }
  }
}

//-----------------------------------------------------------------------------

void WindowMain::UpdateBreakpointsInLuaDebugger()
{
  m_luaThread->ClearBreakpoints();

  int   cnt = ui->scintilla->lines();
  int   line;

  for (line = 0; line < cnt; line++)
  {
    if (ui->scintilla->markersAtLine(line) & (1 << m_scintillaMarkerBreakpoint))
      m_luaThread->SetBreakpoint(line + 1);
  }
}

//-----------------------------------------------------------------------------

QFont WindowMain::GetFixedFont(int fontSz) const
{
  QFontDatabase fontDatabase;
  QStringList   fontList = fontDatabase.families(QFontDatabase::Latin);

  bool  done;
  int   i;

  do
  {
    done = true;

    for (i = 0; i < fontList.size(); i++)
    {
      QString family = fontList.at(i);
      bool    remove = false;

      if (!fontDatabase.isFixedPitch(family))
        remove = true;

      if (!fontDatabase.pointSizes(family).contains(fontSz))
        remove = true;

      if (remove)
      {
        fontList.removeAt(i);
        done = false;
        break;
      }
    }
  }
  while (!done);

  QString fontName;

  for (i = 0; i < fontList.size(); i++)
  {
    if (fontList.at(i).contains("monospace", Qt::CaseInsensitive))
    {
      fontName = fontList.at(i);
      break;
    }
  }

  if (fontName.isEmpty())
  {
    for (i = 0; i < fontList.size(); i++)
    {
      if (fontList.at(i).contains("courier", Qt::CaseInsensitive))
      {
        fontName = fontList.at(i);
        break;
      }
    }
  }

  if (fontName.isEmpty())
  {
    if (fontList.size() > 0)
      fontName = fontList.at(0);
    else
      fontName = "Courier";
  }

  QFont font(fontName);

  font.setStyleHint(QFont::TypeWriter);
  font.setPointSize(fontSz);
  font.setFixedPitch(true);

  return font;
}

//-----------------------------------------------------------------------------
