/*
  Copyright (C) 2012  Torsten Brischalle

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/
*/

#ifndef WINDOWMAIN_H
#define WINDOWMAIN_H

//-----------------------------------------------------------------------------

#include <QMainWindow>
#include <QFileDialog>
#include <QSharedPointer>
#include <QTreeWidget>
#include <QGraphicsScene>
#include <QTableWidget>

#include <iostream>
#include <stdexcept>
#include <map>

#include "luathread.h"
#include "qconsolestreambuf.h"
#include "windowfunctionhelp.h"
#include "windowabout.h"
#include "qcustomplot.h"
#include "findreplacedialog.h"
#include "finddialog.h"

//-----------------------------------------------------------------------------

#define MAX_RECENT_FILES              10
#define DEBUG_INFO_MAX_ARRAY_ITEMS    50

//-----------------------------------------------------------------------------

namespace Ui
{
  class WindowMain;
}

//-----------------------------------------------------------------------------

class WindowMain
  : public QMainWindow
{
  Q_OBJECT

public:
  explicit WindowMain(QWidget *parent = 0);
  ~WindowMain();

  void ProgramStarted();

private:

  typedef QMap <QString, QCPGraph *> CurveList;

  class Diagram
  {
  public:
    QCustomPlot *m_plot;
    int         m_curveColorIndex;
    CurveList   m_curveList;
  };

  typedef QMap <QString, Diagram> DiagramList;

  DiagramList     m_diagramList;

  QList <QColor>  m_curveColorList;

  QCustomPlot     *m_currentDiagramContextMenu;

  std::basic_streambuf <char> *m_oldStreamBufOut;
  std::basic_streambuf <char> *m_oldStreamBufErr;
  QConsoleStreamBuf           *m_consoleStreamBufOut;
  QConsoleStreamBuf           *m_consoleStreamBufErr;

  QStringList m_recentFiles;
  QString     m_currentFile;

  Ui::WindowMain *ui;

  WindowFunctionHelp  m_winFuncHelp;
  WindowAbout         m_winAbout;
  FindReplaceDialog   *m_findReplaceDialog;
  FindDialog          *m_findDialog;

  QSharedPointer <LuaThread> m_luaThread;

  QList <int> m_debugInfoWidthDebugOn, m_debugInfoWidthDebugOff;

  QMap <QString, bool>  m_globalVarFilter;

  QTreeWidgetItem *m_globalVarsItem;
  QTreeWidgetItem *m_localVarsItem;

  QGraphicsScene  m_graphicsScene;
  QColor          m_graphicsSceneColor;
  QColor          m_graphicsSceneFillColor;

  int             m_scintillaMarkerBreakpoint;
  int             m_scintillaMarkerHalted;

  QFont           m_font;

  QActionGroup    m_actionGroupGuiStyle;

  enum GuiState { GS_STOPPED, GS_RUNNING, GS_HALTED };

  QFont GetFixedFont(int fontSz) const;

  void SetGuiState(GuiState guiState);

  void AddToRecentFiles(const QString & fileName);
  void UpdateRecentFilesMenu();

  bool LoadFile(const QString & fileName);
  void SaveFile(QString fileName);

  void UpdateWindowTitle();

  bool ContinueActionIfModified();
  void ShowDebugInfo();
  void ShowDebugInfoCallStack();
  void ShowDebugInfoVariables();

  void SaveDiagramDataAsCsv(const QString & fileName, QCustomPlot *plot);
  void SaveTableDataAsCsv(const QString &fileName, QTableWidget *table);

  void UpdateBreakpointsInLuaDebugger();

public slots:
  void LuaStarted();
  void LuaStopped();
  void LuaHalted();
  void LuaContinued();

  void modificationChanged(bool changed);

  void SlotTableClear();
  void SlotTableSetSize(int rows, int columns);
  void SlotTableSetValue(int row, int column, const QString & value);
  void SlotTableSetValuesInRow(int row, const QList <QString> & valueList);
  void SlotTableSetValuesInColumn(int column, const QList <QString> & valueList);
  void SlotTableSaveCsv(const QString & fileName);

  void SlotDiaDiagramNew(const QString & diagramName);
  void SlotDiaDiagramDelete(const QString & diagramName);
  void SlotDiaDiagramSaveImg(const QString & diagramName, const QString & fileName);
  void SlotDiaDiagramSaveCsv(const QString & diagramName, const QString & fileName);
  void SlotDiaDiagramDeleteAll();
  void SlotDiaCurveNew(const QString & diagramName, const QString & curveName, int yAxis);
  void SlotDiaCurveDelete(const QString & diagramName, const QString & curveName);
  void SlotDiaCurveDeleteAll(const QString & diagramName);
  void SlotDiaCurveClearData(const QString & diagramName, const QString & curveName);
  void SlotDiaCurveSetData(const QString & diagramName, const QString &curveName, const QVector <double> & x, const QVector <double> & y);
  void SlotDiaCurveAppendData(const QString & diagramName, const QString &curveName, double x, double y);

  void SlotVecGfxClear();
  void SlotVecGfxSetColor(int red, int green, int blue);
  void SlotVecGfxSetColor(const QString & name);
  void SlotVecGfxSetFillColor(int red, int green, int blue);
  void SlotVecGfxSetFillColor(const QString & name);
  void SlotVecGfxDrawLine(qreal x1, qreal y1, qreal x2, qreal y2);
  void SlotVecGfxDrawCircle(qreal x, qreal y, qreal r);
  void SlotVecGfxDrawEllipse(qreal x, qreal y, qreal rx, qreal ry);
  void SlotVecGfxDrawRect(qreal x, qreal y, qreal w, qreal h);
  void SlotVecGfxDrawPolygon(const QPolygonF & polygon);
  void SlotVecGfxFillCircle(qreal x, qreal y, qreal r);
  void SlotVecGfxFillEllipse(qreal x, qreal y, qreal rx, qreal ry);
  void SlotVecGfxFillRect(qreal x, qreal y, qreal w, qreal h);
  void SlotVecGfxFillPolygon(const QPolygonF & polygon);
  void SlotVecGfxDrawText(qreal x, qreal y, const QString & text);
  void SlotVecGfxSaveImg(const QString & fileName, int widht, int height);
  void SlotVecGfxSaveCsv(const QString & fileName);

private slots:
  void slotContextMenuRequest(QPoint pos);
  void slotOpenRecentFile();
  void slotZoomToFit();
  void slotDiagramSaveImage();
  void slotDiagramSaveCsv();
  void slotDebugInfoVariablesItemExpanded(QTreeWidgetItem *item);
  void slotMarginClicked(int margin, int line, Qt::KeyboardModifiers state);
  void slotGuiStyle();
  void on_actionStop_triggered();
  void on_actionRun_triggered();
  void on_actionLoad_triggered();
  void on_actionSave_triggered();
  void on_actionSave_As_triggered();
  void on_actionFunction_help_triggered();
  void on_actionAbout_triggered();
  void on_actionSave_image_triggered();
  void on_actionSave_data_triggered();
  void on_actionDebug_triggered();
  void on_actionHalt_triggered();
  void on_actionContinue_triggered();
  void on_actionStepIn_triggered();
  void on_actionClear_all_breakpoints_triggered();
  void on_actionStepOut_triggered();
  void on_actionStepOver_triggered();
  void on_actionNew_triggered();
  void on_actionReset_Lua_triggered();
  void on_actionFind_and_replace_triggered();
  void on_actionFind_triggered();
  void on_actionIndent_triggered();
  void on_actionUnindent_triggered();
  void on_actionSave_image_2_triggered();
  void on_actionSave_SVG_triggered();
  void on_actionUndo_triggered();
  void on_actionRedo_triggered();

protected:
  virtual void closeEvent(QCloseEvent *event);
};

//-----------------------------------------------------------------------------

#endif // WINDOWMAIN_H
